# coding: utf-8
__all__ = ['tag', 'tagged', 'untagged', 'JsonTagError', 'TaggedValue']

import types
_tag_class_dict = {
    int: 'integer',
    float: 'number',
    bool: 'boolean',
    dict: 'object',
    unicode: 'string',
    str: 'binary',
    types.NoneType: 'null',
    list: 'array',
    tuple: 'array',
}

_reserved = set(_tag_class_dict.values())

class JsonTagError(Exception):
    pass

class TaggedValue(object):
    """type tag 付きのスカラー値用のクラス。
    {$tag:foo, $val:1} のようなオブジェクトは、あくまでもスカラー値として扱いたい。
    そのため、タグ付きのデータはすべて {} や dict などの組み込み型ではなく
    専用のタグ型を用意し、タグ型を int などのサブクラスとして扱うことで対処する。
    この TaggedValue クラスは単体では使用せず、 to_tagged_value 関数経由で使用する。
    実際には TaggedValue と $val に格納するデータを多重継承したクラスを使用することになる。
    """
    @property
    def value(self):
        return self

    @property
    def tag(self):
        return self.__tag__

    def to_dict(self):
        return {'$tag': self.tag, '$val': self}

def tagged(tagname, val):
    if tagname in _reserved:
        raise JsonTagError(u'%s は予約語です' % tagname)
    cls = type(val)
    class _TaggedValue(TaggedValue, cls):
        def __getattr__(self, name):
            return cls.__getattr__(self, name)

        def __str__(self):
            return u'@%s %s' % (self.tag, cls.__str__(self))

    r = _TaggedValue(val)
    r.__tag__ = tagname
    return r

def tag(val, explicit=False):
    try:
        return val.tag
    except:
        if explicit:
            raise JsonTagError(u'明示的なタグが指定されていません')
        t = type(val)
        try:
            return _tag_class_dict[t]
        except:
            raise JsonTagError(u'対応する JSON 型が存在しません： %s' % (str(t)))

def untagged(val, explicit=False):
    try:
        return val.value
    except:
        if explicit:
            raise JsonTagError(u'明示的なタグが指定されていません')
        return val


