# coding: utf-8
from caty.jsontools.path.JsonPathLexer import Lexer
from caty.jsontools.path.JsonPathParser import Parser, undefined, JsonPathError
from caty.util import merge_dict, try_parse

__all__ = ['to_decl_style']

def to_decl_style(fo):
    l = Lexer(fo)
    p = JsonValidatorBuilder(l)
    p.parse_jsonpath()
    return p.build_comeplete_tree()

class JsonValidatorBuilder(Parser):
    u"""JSON パス式スキーマのトークン列からスキーマオブジェクトを構築する。
    パス式によるスキーマの表現の場合、パスの途中ノードのスキーマは推論できる。
    $.foo.#.bar: string という式からは以下のような型が導出される。

    type $ = object {
        foo: array[{
            bar: string,
            *: any
            }]
    }

    また、 $.foo.0: integer; $.foo.#: string からは

    type $ = {
        foo:[integer, string*]
    }

    という型が導出される。"""

    def __init__(self, *args, **kwds):
        Parser.__init__(self, *args, **kwds)
        self.path_list = []
        self.tree_list = []
        self.schema_tree = {}
        self.optname = None

    def build_comeplete_tree(self):
        if self.path_list != []:
            self._build_tree()
        tree = reduce(lambda x, y:merge_dict(x, y, is_same_schema), self.tree_list)
        return convert_schema(tree)

    def _build_tree(self):
        tree = reduce(lambda x, y: {y: x}, reversed(self.path_list))
        self.tree_list.append(tree)

    def _root(self, token):
        if self.path_list != []:
            self._build_tree()
            self.path_list = []
        self.path_list.append(token.getText())

    def _path(self, token):
        if self.path_list == []:
            raise JsonPathError("Syntax Error")
        self.path_list.append(token.getText())

    def _schema_type(self, token):
        if self.path_list == []:
            raise JsonPathError("Syntax Error")
        self.path_list.append(token.getText())

def is_same_schema(a, b):
    u"""パス式の推論上での衝突判定関数。
    次のようなパス式でスキーマを書いた場合、処理の都合上衝突が起こってしまう。

    $.foo: array
    $.foo.#: string

    この場合、 $.foo.#: string から $.foo が array[string*] だと導出され、
    array は array[string*] を包含するため、このスキーマは妥当なスキーマとなる。
    それに対して、次のケースではエラーとなる。

    $.bar: array[string*]
    $.bar.#: integer

    array[string*] と array[intger*] はまったく包含関係にないので、これはエラーである。
    ただし現在の実装では厳密な型検査は行っておらず、上記のケースでは array[intger*] に修正される。
    
    次のようなケースは常にエラーとなる。

    $.bar: string
    $.bar: string
    """
    if isinstance(a, dict):
        if b.strip().startswith('array') and is_array(a):
            return 0
        if b.strip().startswith('object') and is_object(a):
            return 0
        return -1
    elif isinstance(b, dict):
        if a.strip().startswith('array') and is_array(b):
            return 1
        if a.strip().startswith('object') and is_object(b):
            return 1
        return -1
    else:
        return -1

def is_array(d):
    for k in d.keys():
        if k != '#' and try_parse(int, k) == None:
            return False
    return True

def is_object(d):
    return not is_array(d)

def convert_schema(tree, root=True):
    u"""パス式から変換された木（実際は辞書）を文字列形式に変換する。
    引数の tree は単一のスキーマの場合もある。
    """
    def _handle_array(dct):
        def _convert(v):
            if isinstance(v, dict):
                return 'object {%s}' % convert_schema(v, False)
            else:
                return v
        if '#' in dct and len(dct) == 1: # '#' だけが指定された場合
            v = dct['#']
            return ('array [%s*]' % _convert(v))
        else: # インデックスも指定されている場合
            keys = filter(lambda x: x, map(lambda x:try_parse(int, x), dct.keys()))
            keys.sort()
            l = []
            for k in map(str, keys):
                l.append(_convert(dct[k]))
            if '#' in dct: # '#' が指定された場合
                l.append(_convert(dct['#'])+'*')
            return 'array [%s]' % ', '.join(l)

    def _convert():
        if not isinstance(tree, dict):
            yield tree
        else:
            lval = ''
            for k, v in tree.iteritems():
                if k == '$':
                    lval = 'type $ = '
                else:
                    lval = '"%s" : ' % k
                if isinstance(v, basestring):
                    yield lval + v
                elif isinstance(v, dict):
                    if is_array(v):
                        yield lval + _handle_array(v)
                    else:
                        yield lval + ('object {%s}' % convert_schema(v, False))

    return ', '.join(list(_convert())) + (';' if root else '')

