#coding: utf-8
from caty.jsontools.schema.schematypes import NullSchema
from caty.util.optionparser import OptionParser, OptionParseFailed
from itertools import *
from functools import *

class ProfileContainer(object):
    u"""複数の CommandProfile をまとめるクラス。
    オプション引数位置固定引数で CommandProfile を決定する。
    入出力の型がオーバーロードされている場合、
    その後の処理は CommandProfile で行う。
    """
    def __init__(self, uri):
        self.uri = uri
        self.profiles = []

    def add_profile(self, profile):
        self.profiles.append(profile)

    def determine_profile(self, args):
        lasterror = None
        for p in self.profiles:
            try:
                args, opts = p.parse_option(args)
                return args, opts, p
            except Exception, e:
                lasterror = e
        raise lasterror
    
    def get_command_class(self):
        abspath = self.uri.split(':')[-1]
        modname, clsname = abspath.rsplit('.', 1)
        mod = __import__(modname, fromlist=[clsname])
        return getattr(mod, clsname)
    

class CommandProfile(object):
    u"""コマンドの引数や型の宣言を定義するクラス。
    一つのコマンドが複数の引数タイプなどでオーバーロードされるため、
    このクラス経由でコマンド確定時の入出力のスキーマを決定する。
    また、以下のようにコマンドの型宣言には複数の入出力のプロファイルがありうる。

        command foo {} [] :: null -> string,
                             string -> string
                             uses some_facility;

    この場合入力が null のものがデフォルトで採用される。
    また、プロファイルが一つしかない場合は常にそのプロファイルを利用する。
    """
    def __init__(self, opts_schema, args_schema, declobj):
        self.opts_schema = opts_schema
        self.args_schema = args_schema
        self.declobj = declobj
        self.in_schema = None
        self.out_schema = None
        if len(declobj.profiles) == 1:
            self.in_schema, self.out_schema = declobj.profiles[0]
        else:
            for i, o in declobj.profiles:
                if isinstance(i, NullSchema):
                    self.in_schema = i
                    self.out_schema = o
                    break
    
    def convert(self, value):
        u"""型変換処理。
        Caty のコマンドオブジェクトは複数のプロファイルを持ちうるため、
        このメソッドでは取りうるプロファイルについて総ざらいに変換処理を行い、
        成功した時点でプロファイルを確定する。
        """
        lasterror = None
        for i, o in self.declobj.profiles:
            try:
                v = i.convert(value)
                self.in_schema = i
                self.out_schema = o
                return v
            except Exception, e:
                lasterror = e
        raise lasterror

    def validate(self, value):
        u"""妥当性検証処理。
        基本的に convert と同様。
        """
        lasterror = None
        for i, o in self.declobj.profiles:
            try:
                i.validate(value)
                self.in_schema = i
                self.out_schema = o
                return
            except Exception, e:
                lasterror = e
        raise e

    def parse_option(self, arguments):
        u"""コマンドライン引数の解析。
        解析結果は辞書とリストのタプルであり、リストの値はスキーマによって変換された結果となる。
        """
        parser = self._make_option_parser()
        opts, args = parser.parse_args(arguments)
        if opts == {}: opts = None
        if args == []: args = None
        return opts, self.args_schema.convert(args)

    def _make_option_parser(self):
        parser = OptionParser()
        already_used = ['h']
        if self.opts_schema.dump(0).strip() in ('null', 'never', 'void'): return parser
        for k, v in self.opts_schema:
            longopt = '--' + k
            shortopt = '-' + dropwhile(lambda x: x in already_used, list(k)).next()
            already_used.append(shortopt[1])
            type = v.dump(0).strip()
            remarks = v.remarks
            if type == 'number':
                type = 'float'
            elif type == 'integer':
                type = 'int'
            if type not in ('int', 'float', 'string', 'boolean'):
                raise Exception('Unsupported type: %s' % type)
            o = partial(parser.add_option, str(shortopt), str(longopt), dest=str(k))
            if type == 'boolean':
                o(action='store_true')
            else:
                o(action='store', type=type)
        return parser

    @property
    def facilities(self):
        u"""使用を宣言されたリソース(=Catyファシリティ)の名前を返す。
        """
        return list(self.declobj.get_all_resource_names())

    def get_command_class(self):
        return self.declobj.get_command_class()

