#coding: utf-8
from caty.jsontools.schema.schematypes import schemas
from caty.jsontools.schema.commandprofile import CommandProfile, ProfileContainer

class Provide(object):
    def __init__(self, names):
        self.exports = names

    def build_schema_tree(self, registrar):
        registrar.exports= self.exports

class SchemaNode(object):
    def __init__(self, name, schema_node, optional=False):
        self.name = name
        self.schema_node = schema_node
        if optional:
            self.schema_node.options['optional'] = optional

    def build_schema_tree(self, registrar):
        registrar.add_schema(self.name, self.schema_node.build(registrar))

class Node(object):
    def __init__(self):
        self.options = {}

    def optional():
        def _get(self):
            return 'optional' in self.options and self.options['optional']
        def _set(self, value):
            self.options['optional'] = value
        return _get, _set
    optional = property(*optional())

class ScalarNode(Node):
    def __init__(self, schema_name):
        Node.__init__(self)
        self.schema_name = schema_name
        self.underlyingtypes = None

    def build(self, registrar):
        schema = registrar.get_schema(self.schema_name)
        udt = []
        if self.underlyingtypes:
            for t in self.underlyingtypes:
                udt.append(t.build(registrar))
        ns = schema.create(self.options if self.options else {}, udt)
        return ns

class UnionNode(Node):
    def __init__(self, node1, node2):
        Node.__init__(self)
        self.node1 = node1
        self.node2 = node2

    def build(self, registrar):
        s1 = self.node1.build(registrar)
        s2 = self.node2.build(registrar)
        s = s1 | s2
        if 'optional' in self.options:
            s.optional = self.options['optional'] 
        return s

class IntersectionNode(Node):
    def __init__(self, node1, node2):
        Node.__init__(self)
        self.node1 = node1
        self.node2 = node2

    def build(self, registrar):
        s1 = self.node1.build(registrar)
        s2 = self.node2.build(registrar)
        s = s1 & s2
        s.optional = self.options['optional']
        return s

class TaggedNode(Node):
    def __init__(self, tag, node):
        self.tag = tag
        self.schema_node = node

    def build(self, registrar):
        schema = self.schema_node.build(registrar)
        return schema.to_tagged(self.tag)

    def optional():
        def _get(self):
            return 'optional' in selaf.schema_node.options and self.schema_node.options['optional']
        def _set(self, value):
            self.schema_node.options['optional'] = value
        return _get, _set
    optional = property(*optional())

class EnumNode(Node):
    def __init__(self, enum):
        Node.__init__(self)
        self.enum = enum

    def build(self,registrar):
        schema = registrar.get_schema('enum')
        return schema.create({'optional': self.optional}, self.enum)

class ObjectNode(ScalarNode):
    def __init__(self, items=None, wildcard=None):
        ScalarNode.__init__(self, 'object')
        self.items = items if items else {}
        self.wildcard = wildcard

    def build(self, registrar):
        items = {}
        for k, v in self.items.iteritems():
            items[k] = v.build(registrar)
        schema = registrar.get_schema(self.schema_name)
        schema = schema.create(self.options if self.options else {}, None)
        if self.wildcard:
            schema.wildcard = self.wildcard
        schema.schemas = items
        return schema

class CommandNode(object):
    def __init__(self, name, patterns, uri, doc):
        self.name = name
        self.patterns = patterns
        self.uri = [u for l, u in uri if l == 'python:'][0]
        self.doc = doc

    def build_schema_tree(self, registrar):
        pc = ProfileContainer(self.uri)
        for p in self.patterns:
            p.build(registrar)
            pc.add_profile(CommandProfile(p.opt_schema, p.arg_schema, p.decl))
        registrar.add_command(self.name, pc)

class CallPattern(object):
    def __init__(self, opts, args, decl):
        self.opts = opts
        self.args = args
        self.decl = decl

    def build(self, registrar):
        if self.opts:
            self.opt_schema = self.opts.build(registrar)
        else:
            self.opt_schema = schemas['null']
        if self.args:
            self.arg_schema = self.args.build(registrar)
        else:
            self.arg_schema = schemas['null']
        self.decl.build(registrar)

class CommandDecl(object):
    u"""コマンドの宣言部分の構造体。以下の要素それぞれのリストを持つ構造体クラスである。
    型宣言 type -> type
    大域脱出宣言 throws, breaks
    資源使用宣言 uses, reads, updates
    コマンドURI参照 refers
    """
    def __init__(self, profiles, jump, resource):
        self.uri = '' # 後で挿入される
        self.profiles = profiles
        self.jump = jump if isinstance(jump, list) else [jump]
        self.resource = resource if isinstance(resource, list) else [resource]

    def get_all_resource_names(self):
        for n, rs in self.resource:
            for r in rs:
                yield r

    def build(self, registrar):
        p = []
        for i, o in self.profiles:
            p.append((i.build(registrar), o.build(registrar)))
        self.profiles = p

        j = []
        for t, ls in self.jump:
            l = []
            for node in ls:
                l.append(node.build(registrar))
            j.append(l)

        self.jump  = j

