#coding: utf-8
from StringIO import StringIO
from caty.jsontools.schema.language.ast import *
from caty.jsontools.schema.language.JsonSchemaLexer import (Lexer,
                                                   STRING,
                                                   SYMBOL,
                                                   COLON,
                                                   LPAREN,
                                                   RPAREN,
                                                   LBRACE,
                                                   RBRACE,
                                                   LBRACKET,
                                                   RBRACKET,
                                                   COMMA,
                                                   EQUAL,
                                                   DELIM,
                                                   TAG,
                                                   OR,
                                                   AND,
                                                   ASTER,
                                                   OPTIONAL,
                                                   NUMBER)

RESERVED = [
    'object',
    'array',
    'list',
    'tuple',
    'string',
    'integer',
    'number',
    'any',
    'enum',
    'null',
    'never',
    'bag',
    'multi',
    'bool',
]

class SchemaSyntaxError(Exception):
    def __init__(self, *args):
        if len(args) == 1:
            self.token = args[0]
        elif len(args) == 2:
            self.token = (args[0], args[1])
        else:
            self.token = args

    def __str__(self):
        return 'JSON schema syntax error: ' + str(self.token)

class StateMachine(object):
    u"""状態機械の抽象クラス。
    すべてのサブクラスは _next と build_ast の二つのメソッドをオーバーライドしなければならない。
    また、任意で _init メソッドをオーバーライドし、そこに独自の初期化コードを書いても良い。
    _next は int -> str -> StateMachine|None|SchemaSyntaxError という型を持ち、
    その状態で受理できないトークンを受け取ったら None を返し、親の状態に処理を戻す。
    いずれの親でもトークンを受理できなかった時にエラーとなる。
    また、即座に文法エラーとしたい場合には SchemaSyntaxError を送出しても良い。
    """
    def __init__(self, parent=None):
        self.parent = parent
        self.child = None
        self.completed = False
        self._init()

    def next(self, token):
        #print self, token
        tp = token.getType()
        val = token.getText()
        try:
            n = self._next(tp, val)
        except SchemaSyntaxError, e:
            raise SchemaSyntaxError(token)
        else:
            if n == None:
                if self.parent:
                    return self.parent.next(token)
            else:
                if n != self and self.parent != n:
                    self.child = n
                return n
            raise SchemaSyntaxError(token)

    def _init(self):
        pass

    def _next(self, tp, val):
        raise NotImeplementedError

    def build_ast(self):
        raise NotImeplementedError

class InitialState(StateMachine):
    u"""初期状態。
    この時点ではスキーマの型宣言とコマンドの型宣言の両者を受け取れる。
    type foo = string;
    type bar = object {
        ...
    };
    command cmd = object {
        "input": ...,
        "output": ...,
        "args": ...,
        "options": ...
    };
    """
    def _next(self, tp, val):
        if tp == 4:
            if val == 'type':
                return TypeState(self)
            elif val == 'command':
                return CommandState(self)

class TypeState(StateMachine):
    u"""スキーマ型宣言を受理する状態機械。
    """
    def _init(self):
        self.name = ''
        self.optional = False
        self.paren_count = 0

    def _next(self, tp, val):
        if tp == SYMBOL:
            if self.name == '':
                if val in RESERVED:
                    raise SchemaSyntaxError('"%s" is reserved' % val)
                self.name = val
                return self
        else:
            if self.name == '':
                return 
            elif tp == EQUAL:
                return SchemaState(self)
            elif tp == OR and self.child and not isinstance(self.child, UnionState):
                if isinstance(self.child, SchemaState):
                    if self.child.found_paren >= 1:
                        self.paren_count = self.child.found_paren 
                union = UnionState(self, self.paren_count)
                self.child = union
                return union
            elif tp == AND and self.child and not isinstance(self.child, InitialState):
                if isinstance(self.child, SchemaState):
                    if self.child.found_paren >= 1:
                        self.paren_count = self.child.found_paren 
                inter = IntersectionState(self, self.paren_count)
                self.child = inter
                return inter
            elif tp == OPTIONAL:
                self.optional = True
                return self
            elif tp == DELIM:
                self.completed = True
                return self

    def build_ast(self):
        return RootNode(self.name, self.child.build_ast(), self.optional)

class SchemaState(StateMachine):
    u"""スキーマ型に応じて状態を振り分けるための状態機械。
    直後に現れるトークンが object であれば object 構築状態へ、
    それ以外のシンボルであればスカラー型へ、
    文字列トークンであればタグ付き型へと遷移する。
    """
    
    def _init(self):
        self.found_paren = 0

    def _next(self, tp, val):
        if tp == 11:
            if val == 'object':
                return ObjectState(self)
        if tp == SYMBOL:
            if val == 'enum':
                return EnumState(self)
            else:
                s = ScalarState(self)
                s.type_name = val
                return s
        elif tp == TAG:
            t = TaggedSchemaState(self)
            t.tag = val
            self.child = t
            return t
        elif tp == LPAREN and not self.child:
            self.found_paren += 1
            return self
        elif tp == RPAREN and self.found_paren > 0:
            self.found_paren -= 1
            return self

    def build_ast(self):
        return self.child.build_ast()

class UnionState(SchemaState):
    def __init__(self, parent, paren_count = 0):
        SchemaState.__init__(self, parent)
        self.another = parent.child
        self.found_paren = paren_count

    def _next(self, tp, val):
        r = SchemaState._next(self, tp, val)
        if r:
            if r != self:
                self.child = r
            return r
        if tp == OR and self.child and not isinstance(self.child, UnionState):
            union = UnionState(self, self.found_paren)
            self.child = union
            return union

    def build_ast(self):
        return UnionNode(self.another.build_ast(), self.child.build_ast())

class IntersectionState(UnionState):
    def _next(self, tp, val):
        r = SchemaState._next(self, tp, val)
        if r:
            if r != self:
                self.child = r
            return r
        if tp == AND and self.child and not isinstance(self.child, IntersectionState):
            inter = IntersectionState(self, self.found_paren)
            self.child = inter
            return inter

    def build_ast(self):
        return IntersectionNode(self.another.build_ast(), self.child.build_ast())

class ScalarState(StateMachine):
    u"""object でない型のスキーマを受理する状態機械。
    以下のような宣言に対して働く。
    type foo = array (maxLength=10) [int, string*];
    このとき () 内のトークンはすべてオプションであり、
    同様に [] 内のトークンは underlying type として扱う。
    配列、リストなどは underlying type とは言わないのだが、
    あくまでも構文要素の名前なので気にしないこととする。
    """
    def _init(self):
        self.options = {}
        self.underlyingtypes = []
        self.parsing_opt = False
        self.parsing_udt = False
        self.curtoken = None
        self.found_eq = False
        self.match_comma = False
        self.type_name = ''
        self.wait_union = False

    def _next(self, tp, val):
        if tp in (LPAREN, RPAREN, LBRACKET, RBRACKET):
            if not self._switch_parse_mode(tp, val):
                return
            else:
                return self
        if self.parsing_opt:
            return self.opt_parse(tp, val)
        elif self.parsing_udt:
            return self.udt_parse(tp, val)

    def opt_parse(self, tp, val):
        if self.match_comma:
            if tp == COMMA:
                self.match_comma = False
                return self
            else:
                return
        if tp == SYMBOL and self.parsing_opt:
            self.curtoken = val
            return self
        elif tp == EQUAL and self.parsing_opt and self.curtoken:
            self.found_eq = True
            return self
        elif tp in (STRING, NUMBER, SYMBOL) and self.parsing_opt and self.found_eq and self.curtoken:
            if tp == STRING:
                v = val
            elif tp == NUMBER:
                v= eval(val)
            else:
                if val in ('true', 'false'):
                    v = eval(val.capitalize())
            self.options[self.curtoken] = v
            self.curtoken = None
            self.found_eq = False
            self.match_comma = True
            return self

    def udt_parse(self, tp, val):
        if self.match_comma:
            if tp == COMMA:
                self.match_comma = False
                return self
            else:
                if tp not in (OPTIONAL, ASTER, OR):
                    return
                else:
                    self.match_comma = False
        if tp == ASTER:
            if 'repeat' not in self.options:
                self.options['repeat'] = True
                self.match_comma = True
                return self
        else:
            if 'repeat' in self.options:
                #繰り返し指定子が出現したら、それ以降にトークンが出現してはいけない
                raise SchemaSyntaxError(tp, val)
        if tp == SYMBOL:
            if self.wait_union:
                self.underlyingtypes[-1] = UnionNode(self.underlyingtypes[-1], ScalarNode(val))
            else:
                self.underlyingtypes.append(ScalarNode(val))
            self.match_comma = True
            return self
        elif tp == OPTIONAL:
            if self.underlyingtypes and not self.underlyingtypes[-1].optional:
                #直前に出現した型にオプショナル属性を付ける
                self.underlyingtypes[-1].optional = True
                self.match_comma = True
                return self
        elif tp == OR:
            if self.underlyingtypes and not self.wait_union:
                self.wait_union = True
                return self

    def _switch_parse_mode(self, tp, val):
        if tp == LPAREN:
            if not self.parsing_opt and not self.parsing_udt:
                self.parsing_opt = True
                self._reset()
            else:
                return False
        elif tp == RPAREN:
            if self.parsing_opt and not self.parsing_udt:
                self.parsing_opt = False
                self._reset()
            else:
                return False
        elif tp == LBRACKET:
            if not self.parsing_udt and not self.parsing_opt:
                self.parsing_udt = True
                self._reset()
            else:
                return False
        elif tp == RBRACKET:
            if self.parsing_udt and not self.parsing_opt:
                self.parsing_udt = False
                self._reset()
            else:
                return False
        return True

    def _reset(self):
        self.curtoken = None
        self.found_eq = False
        self.match_comma = False

    def build_ast(self):
        s = ScalarNode(self.type_name)
        s.options = self.options
        s.underlyingtypes = self.underlyingtypes
        return s

class EnumState(StateMachine):
    def _init(self):
        self.enum = []
        self.started = False
        self.ended = False
        self.wait_comma = False

    def _next(self, tp, val):
        if tp == LBRACKET:
            if not self.started:
                self.started = True
                return self
        elif tp == RBRACKET:
            if not self.ended:
                self.ended = True
                return self
        elif (tp == NUMBER or tp == STRING) and self.started and not self.ended and not self.wait_comma:
            if tp == NUMBER:
                if not '.' in val:
                    self.enum.append(int(val))
            else:
                self.enum.append(val)
            t = self.enum[0].__class__
            if not all(map(lambda x: x.__class__ == t, self.enum)):
                raise SchemaSyntaxError(val)
            self.wait_comma = True
            return self
        elif tp == COMMA and self.wait_comma:
            self.wait_comma = False
            return self

    def build_ast(self):
        return EnumNode(self.enum)

class TaggedSchemaState(SchemaState):
    u"""タグ付きの値を処理する。
    """
    def _init(self):
        self.tag = ''

    def build_ast(self):
        return TaggedNode(self.tag, SchemaState.build_ast(self))

class ObjectState(ScalarState):
    u"""object {
    }
    """
    def _init(self):
        ScalarState._init(self)
        self.child_nodes = {}
        self.started = False

    def _next(self, tp, val):
        r = ScalarState._next(self, tp, val)
        if r: return r
        if tp == LBRACE:
            if self.started:
                return
            self.started = True
            self.child = MemberState(self)
            return self.child
        elif tp == COMMA:
            self.child_nodes[self.child.name] = self.child
            self.child = MemberState(self)
            return self.child
        elif tp == RBRACE:
            if self.child.name:
                self.child_nodes[self.child.name] = self.child
            return self.parent

    def build_ast(self):
        o = ObjectNode()
        for k, v in self.child_nodes.iteritems():
            if k == '*':
                o.wildcard = v.build_ast()
            else:
                o.items[k] = v.build_ast()
        return o

class MemberState(StateMachine):
    def _init(self):
        self.name = ''
        self.optional = False

    def _next(self, tp, val):
        if tp == STRING and self.name == '':
            self.name = val
            return self
        else:
            if tp != ASTER and self.name == '':
                return 
            elif tp == COLON:
                return SchemaState(self)
            elif tp == OR and self.child and not isinstance(self.child, UnionState):
                union = UnionState(self)
                self.child = union
                return union
            elif tp == AND and self.child and not isinstance(self.child, InitialState):
                inter = IntersectionState(self)
                self.child = inter
                return inter
            elif tp == ASTER and self.name == '':
                self.name = val
                return self
            elif tp == OPTIONAL:
                self.optional = True
                return self.parent

    def build_ast(self):
        ast = self.child.build_ast()
        if self.optional:
            ast.options['optional'] = True
        return ast

class CommandState(StateMachine):
    u"""コマンド型宣言を受理する状態機械。
    """
    def _init(self):
        self.name = ''

    def _next(self, tp, val):
        if tp == SYMBOL:
            if self.name == '':
                self.name = val
                return self
        else:
            if self.name == '':
                return 
            elif tp == EQUAL:
                return CommandSchemaState(self)
            elif tp == DELIM:
                self.completed = True
                return self

    def build_ast(self):
        return CommandNode(self.name, self.child.build_ast())

class CommandSchemaState(StateMachine):
    def _next(self, tp, val):
        if tp == SYMBOL:
            if val == 'object':
                o = CommandObjectState(self)
                self.child = o
                return o

    def build_ast(self):
        return self.child.build_ast()

class CommandObjectState(ObjectState):
    def build_ast(self):
        o = ObjectNode()
        for k, v in self.child_nodes.iteritems():
            if k in ('input', 'output', 'args', 'options'):
                o.items[k]  = v.build_ast()
            else:
                raise SchemaSyntaxError(k)
        return o

def parse(fo):
    lexer = Lexer(fo)
    state = InitialState()
    for t in lexer:
        state = state.next(t)
        if state.completed:
            yield state.build_ast()
            state = InitialState()
    if state.completed:
        yield state.build_ast()

