# coding: utf-8
import sys
import os
from StringIO import StringIO
sys.path.insert(0, os.path.abspath('../../../../../'))
from caty.jsontools.schema.language.parser import parse as p, SchemaSyntaxError
from caty.jsontools.schema.language.ast import *
from caty.testutil import TestCase, test

def parse(s):
    io =StringIO()
    io.write(s)
    io.seek(0)
    return list(p(io))

class ParserTest(TestCase):
    def test_simple_scalar_type(self):
        s = 'type foo = pkg.mod:Foo;'
        node = parse(s)[0]
        self.assertTrue(isinstance(node, RootNode))
        self.assertEquals(node.name, 'foo')
        self.assertTrue(isinstance(node.schema_node, ScalarNode))
        self.assertEquals(node.schema_node.schema_name, 'pkg.mod:Foo')

    def test_simple_optional1(self):
        s = 'type foo = pkg.mod:Foo?;'
        node = parse(s)[0]
        self.assertTrue(isinstance(node, RootNode))
        self.assertEquals(node.name, 'foo')
        self.assertTrue(isinstance(node.schema_node, ScalarNode))
        self.assertEquals(node.schema_node.schema_name, 'pkg.mod:Foo')

    def test_simple_optional2(self):
        s = 'type foo = (pkg.mod:Foo)?;'
        node = parse(s)[0]
        self.assertTrue(isinstance(node, RootNode))
        self.assertEquals(node.name, 'foo')
        self.assertTrue(isinstance(node.schema_node, ScalarNode))
        self.assertEquals(node.schema_node.schema_name, 'pkg.mod:Foo')

    def test_simple_scalar_type_with_opts(self):
        s = 'type foo = integer(maximum=10, minimun=5);'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, ScalarNode))
        self.assertEquals(node.schema_node.schema_name, 'integer')
        self.assertEquals(node.schema_node.options, {'maximum': 10, 'minimun': 5})

    def test_simple_type_with_udt(self):
        s = 'type foo = array() [int, string, bool];'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, ScalarNode))
        self.assertEquals(node.schema_node.schema_name, 'array')
        self.assertEquals(node.schema_node.options, {})
        self.assertTrue(any(map(lambda a: isinstance(a, ScalarNode), node.schema_node.underlyingtypes)))

    def test_simple_type_repeat(self):
        s = 'type foo = array() [int, string, bool*];'
        node = parse(s)[0]
        self.assertTrue(node.schema_node.options['repeat'])

    def test_simple_union_type(self):
        s = 'type foo = integer? | string;'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, UnionNode))
        self.assertTrue(isinstance(node.schema_node.node1, ScalarNode))
        self.assertTrue(isinstance(node.schema_node.node2, ScalarNode))

        s = 'type foo = (integer | string | bool)?;'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, UnionNode))
        self.assertTrue(isinstance(node.schema_node.node1, ScalarNode))
        self.assertTrue(isinstance(node.schema_node.node2, UnionNode))

    def test_simple_union_with_tag(self):
        s = 'type foo = @i integer | @s string;'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, UnionNode))
        self.assertTrue(isinstance(node.schema_node.node1, TaggedNode))
        self.assertTrue(isinstance(node.schema_node.node2, TaggedNode))

    def test_simple_multiple_decl(self):
        s = 'type foo = string;\ntype bar = integer;'
        node1 = parse(s)[0]
        node2 = parse(s)[1]
        self.assertTrue(isinstance(node1, RootNode))
        self.assertEquals(node1.name, 'foo')
        self.assertTrue(isinstance(node1.schema_node, ScalarNode))
        self.assertEquals(node1.schema_node.schema_name, 'string')

        self.assertTrue(isinstance(node2, RootNode))
        self.assertEquals(node2.name, 'bar')
        self.assertTrue(isinstance(node2.schema_node, ScalarNode))
        self.assertEquals(node2.schema_node.schema_name, 'integer')

    def test_object(self):
        s = """type foo = object {
            "a": string,
            "b": bool,
            *: any
        };
        """
        node = parse(s)[0]
        self.assertTrue(isinstance(node, RootNode))
        self.assertEquals(node.name, 'foo')
        self.assertTrue(isinstance(node.schema_node, ObjectNode))
        n = node.schema_node
        self.assertTrue('a' in n.items)
        self.assertEquals(n.items['a'].schema_name, 'string')
        self.assertTrue('b' in n.items)
        self.assertEquals(n.items['b'].schema_name, 'bool')
        self.assertTrue(isinstance(n.wildcard, ScalarNode))

    def test_nested_object(self):
        s = """type foo = object {
            "a": string,
            "b": object {
                "c": bool?,
                "d": (integer)
            }
        };
        """
        node = parse(s)[0]
        self.assertTrue(isinstance(node, RootNode))
        self.assertEquals(node.name, 'foo')
        self.assertTrue(isinstance(node.schema_node, ObjectNode))
        n = node.schema_node
        self.assertTrue('a' in n.items)
        self.assertTrue('b' in n.items)
        
        self.assertTrue(n.wildcard is None)
        b = n.items['b']
        self.assertTrue('c' in b.items)
        self.assertEquals(b.items['c'].schema_name, 'bool')
        self.assertTrue('d' in b.items)
        self.assertEquals(b.items['d'].schema_name, 'integer')
        
    def test_command(self):
        s = """command foo.bar.buz.Cmd = object {
            "input": string,
            "output": integer,
            "args": array [integer, string?],
            "options": object {
                "debug": bool,
            }
        };
        """
        node = parse(s)[0]
        self.assertTrue(isinstance(node, CommandNode))
        self.assertEquals(node.name, 'foo.bar.buz.Cmd')
        self.assertTrue(isinstance(node.schema_node, ObjectNode))

    def test_enum(self):
        s = 'type foo = enum ["a", "b", "c"];'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, EnumNode))

        s = 'type foo = enum [0, 1, 2];'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, EnumNode))

        s = 'type foo = enum ["a", "b", 1];'
        self.assertRaises(SchemaSyntaxError, parse, s)

    def test_intersection(self):
        s = """type foo = object {
            "a": string
        } & object {"b": integer };
        """
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, IntersectionNode))
        s2 = 'type x = (foo & bar)?;'
        node = parse(s2)[0]
        self.assertTrue(isinstance(node.schema_node, IntersectionNode))

    def test_optional(self):
        s1 = 'type a = array?;'
        self.assertNotRaises(lambda :parse(s1)[0])
        s2 = 'type b = array(maxImtes=10)?;'
        self.assertNotRaises(lambda :parse(s2)[0])
        s3 = 'type c = array(maxImtes=10)[integer*]?;'
        self.assertNotRaises(lambda :parse(s3)[0])

        f1 = 'type d = array ? (maxImtes=10);'
        self.assertRaises(SchemaSyntaxError, lambda :parse(f1)[0])
        f2 = 'type e = array(maxImtes=10)?[integer*];'
        self.assertNotRaises(SchemaSyntaxError, lambda :parse(f2)[0])

if __name__ == '__main__':
    test(ParserTest)

