# coding: utf-8
import sys
import os
from StringIO import StringIO
from caty.jsontools.schema.language.casmparser import parse
from caty.jsontools.schema.language.util import SchemaSyntaxError
from caty.jsontools.schema.language.ast import *
from caty.testutil import TestCase, test

class SchemaParserTest(TestCase):
    def test_simple_scalar_type(self):
        s = 'type foo = pkg.mod:Foo;'
        node = parse(s)[0]
        self.assertTrue(isinstance(node, SchemaNode))
        self.assertEquals(node.name, 'foo')
        self.assertTrue(isinstance(node.schema_node, ScalarNode))
        self.assertEquals(node.schema_node.schema_name, 'pkg.mod:Foo')
        s = 'type foo = pkg:$;'
        node = parse(s)[0]
        s = 'type $ = object {"fuga" :  string};'
        node = parse(s)[0]


    def test_simple_optional1(self):
        s = 'type foo = pkg.mod:Foo?;'
        node = parse(s)[0]
        self.assertTrue(isinstance(node, SchemaNode))
        self.assertEquals(node.name, 'foo')
        self.assertTrue(isinstance(node.schema_node, ScalarNode))
        self.assertEquals(node.schema_node.schema_name, 'pkg.mod:Foo')

    def test_simple_optional2(self):
        s = 'type foo = (pkg.mod:Foo)?;'
        node = parse(s)[0]
        self.assertTrue(isinstance(node, SchemaNode))
        self.assertEquals(node.name, 'foo')
        self.assertTrue(isinstance(node.schema_node, ScalarNode))
        self.assertEquals(node.schema_node.schema_name, 'pkg.mod:Foo')

    def test_simple_scalar_type_with_opts(self):
        s = 'type foo = integer(maximum=10, minimun=5);'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, ScalarNode))
        self.assertEquals(node.schema_node.schema_name, 'integer')
        self.assertEquals(node.schema_node.options, {'maximum': 10, 'minimun': 5})

    def test_simple_type_with_udt(self):
        s = 'type foo = array() [integer, string, boolean];'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, ScalarNode))
        self.assertEquals(node.schema_node.schema_name, 'array')
        self.assertEquals(node.schema_node.options, {})
        self.assertTrue(any(map(lambda a: isinstance(a, ScalarNode), node.schema_node.underlyingtypes)))

    def test_simple_type_repeat(self):
        s = 'type foo = array() [integer, string, boolean*];'
        node = parse(s)[0]
        self.assertTrue(node.schema_node.options['repeat'])
        s = 'type foo = array() [(integer|string)*];'
        node = parse(s)[0]
        self.assertTrue(node.schema_node.options['repeat'])
        s = """type foo = object {"x": string};
        type bar = array() [foo*];"""
        node = parse(s)[1]
        self.assertTrue(node.schema_node.options['repeat'])

    def test_simple_union_type(self):
        s = 'type foo = integer? | string;'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, UnionNode))
        self.assertTrue(isinstance(node.schema_node.node1, ScalarNode))
        self.assertTrue(isinstance(node.schema_node.node2, ScalarNode))

        s = 'type foo = (integer | string | boolean)?;'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, UnionNode))
        self.assertTrue(isinstance(node.schema_node.node1, UnionNode))
        self.assertTrue(isinstance(node.schema_node.node2, ScalarNode))

    def test_simple_union_with_tag(self):
        s = 'type foo = @i integer | @s string;'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, UnionNode))
        self.assertTrue(isinstance(node.schema_node.node1, TaggedNode))
        self.assertTrue(isinstance(node.schema_node.node2, TaggedNode))

    def test_simple_multiple_decl(self):
        s = 'type foo = string;\ntype bar = integer;'
        node1 = parse(s)[0]
        node2 = parse(s)[1]
        self.assertTrue(isinstance(node1, SchemaNode))
        self.assertEquals(node1.name, 'foo')
        self.assertTrue(isinstance(node1.schema_node, ScalarNode))
        self.assertEquals(node1.schema_node.schema_name, 'string')

        self.assertTrue(isinstance(node2, SchemaNode))
        self.assertEquals(node2.name, 'bar')
        self.assertTrue(isinstance(node2.schema_node, ScalarNode))
        self.assertEquals(node2.schema_node.schema_name, 'integer')

    def test_object(self):
        s = """type foo = object {
            "a": string,
            "b": boolean,
            *: any
        };
        """
        node = parse(s)[0]
        self.assertTrue(isinstance(node, SchemaNode))
        self.assertEquals(node.name, 'foo')
        self.assertTrue(isinstance(node.schema_node, ObjectNode))
        n = node.schema_node
        self.assertTrue('a' in n.items)
        self.assertEquals(n.items['a'].schema_name, 'string')
        self.assertTrue('b' in n.items)
        self.assertEquals(n.items['b'].schema_name, 'boolean')
        self.assertTrue(isinstance(n.wildcard, ScalarNode))

    def test_nested_object(self):
        s = """type foo = object {
            "a": string,
            "b": object {
                "c": boolean?,
                "d": (integer)
            }
        };
        """
        node = parse(s)[0]
        self.assertTrue(isinstance(node, SchemaNode))
        self.assertEquals(node.name, 'foo')
        self.assertTrue(isinstance(node.schema_node, ObjectNode))
        n = node.schema_node
        self.assertTrue('a' in n.items)
        self.assertTrue('b' in n.items)
        
        self.assertTrue(n.wildcard is None)
        b = n.items['b']
        self.assertTrue('c' in b.items)
        self.assertEquals(b.items['c'].schema_name, 'boolean')
        self.assertTrue('d' in b.items)
        self.assertEquals(b.items['d'].schema_name, 'integer')
 
        s = """type bar = list [object{"x": string, "y": integer}];"""
        node = parse(s)[0]
        self.assertTrue(isinstance(node, SchemaNode))
        self.assertEquals(node.name, 'bar')
        n = node.schema_node.underlyingtypes[0]
        self.assertTrue(isinstance(n, ObjectNode))
        self.assertTrue('x' in n.items)
        self.assertTrue('y' in n.items)

    def test_enum(self):
        s = 'type foo = enum ["a", "b", "c"];'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, EnumNode))

        s = 'type foo = enum [0, 1, 2];'
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, EnumNode))

        s = 'type foo = enum ["a", "b", 1];'
        self.assertRaises(ValueError, parse, s)

    def test_intersection(self):
        s = """type foo = object {
            "a": string
        } & object {"b": integer };
        """
        node = parse(s)[0]
        self.assertTrue(isinstance(node.schema_node, IntersectionNode))
        s2 = 'type x = (foo & bar)?;'
        node = parse(s2)[0]
        self.assertTrue(isinstance(node.schema_node, IntersectionNode))

    def test_optional(self):
        s1 = 'type a = array?;'
        self.assertNotRaises(lambda :parse(s1)[0])
        s2 = 'type b = array(maxImtes=10)?;'
        self.assertNotRaises(lambda :parse(s2)[0])
        s3 = 'type c = array(maxImtes=10)[integer*]?;'
        self.assertNotRaises(lambda :parse(s3)[0])

        f1 = 'type d = array ? (maxImtes=10);'
        self.assertRaises(SchemaSyntaxError, lambda :parse(f1)[0])
        f2 = 'type e = array(maxImtes=10)?[integer*];'
        self.assertNotRaises(SchemaSyntaxError, lambda :parse(f2)[0])

