#coding: utf-8
import copy
from caty.jsontools.model import TaggedValue

class JsonSchemaError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg.encode('utf-8')

class ISchema(object):
    u"""スキーマインターフェース。
    """

    def create(self, options, underlyingtypes):
        u"""スキーマの生成。
        options は maxLength などのオプション値のマップであり、
        underlyingtypes は string など制約を与えるための型のリストである。
        """
        raise NotImplementedError
        

    def validate(self, value):
        u"""データの検証
        """
        raise NotImplementedError

    def convert(self, value):
        u"""主に Web からの POST データや標準入力からのデータなど、
        データ型を変換しなければならない場合に用いる。
        """
        raise NotImplementedError

    def dump(self, depth=0):
        u"""スキーマ定義のダンプ処理。
        """
        raise NotImplementedError

class SchemaBase(ISchema):
    u"""あらゆるスキーマクラスのベースクラス。
    """
    
    def optional():
        def _get(self):
            if not hasattr(self, '_optional'):
                return False
            return self._optional
        def _set(self, v):
            self._optional = v
        return _get, _set
    optional = property(*optional())
    
    def remarks():
        def _get(self):
            if not hasattr(self, '_remarks'):
                return False
            return self._remarks
        def _set(self, v):
            self._remarks = v
        return _get, _set
    remarks = property(*remarks())

    def __call__(self, *args, **kwds):
        u"""API をより使いやすくするために、関数として呼ばれた場合には新しいインスタンスを返す事にする。
        これは { "a": integer } のような宣言と { "b": integer(maximum=100) } のような宣言を両立させるために用いる。
        あらかじめ intger というシンボルには IntegerSchema のインスタンスを束縛しておき、
        シンボルが関数のように呼び出された場合には、その時の引数を用いて新たなインスタンスを返す。
        このため、基本的にスキーマクラスは不変オブジェクトとして実装されなければならない。
        """
        return self.__class__(*args, **kwds)

    def __or__(self, schema):
        u"""Union 型スキーマの構築。
        本来 Python において | はビット単位の or 演算だが、
        この場合は Union の構築に用いる。
        a | b == UnionSchema(a, b) である。
        """
        return UnionSchema(self, schema)


    def convert(self, value):
        u"""convert メソッドの実装。
        実際の処理は _convert メソッドで行い、その結果を検証した結果を返すことになる。
        デフォルトの _convert の実装では何もせずに値をそのまま返す。
        """
        try:
            r = self._convert(value)
        except Exception, e:
            raise JsonSchemaError(u'型変換時に例外が発生しました。: %s' % str(e))
        self.validate(r)
        return r

    def _convert(self, value):
        return value

    def clone(self):
        return copy.deepcopy(self)

    def to_tagged(self, tag):
        return TagSchema(tag, self)

class UnionSchema(SchemaBase):
    u"""union 型に対応するスキーマ。
    union 型は二つのスキーマによってインスタンス化され、
    それらのスキーマのいずれか一方に適合するデータであれば妥当なデータとみなす。
    両方のスキーマで妥当性検証に失敗したときのみがエラーである。
    """
    def __init__(self, a, b):
        self.__schema_a = a
        self.__schema_b = b

    def validate(self, value):
        u"""二つのスキーマのいずれかで妥当性検証に通るかの処理を行う。
        value が $case プロパティを持っている場合、タグ付き Union と看做す。
        """
        if isinstance(value, TaggedValue):
            self._validate_tagged_union(value)
        else:
            self._validate_normal_union(value)

    def _validate_tagged_union(self, value):
        u"""タグ付きの union に対する処理。
        value に $val プロパティがあればそれを値とし、
        なければ $case を除いたプロパティを値とする。
        """
        tag = value.tag
        if not isinstance(value, dict):
            new_val = value
        else:
            new_val = dict([(k, v) for k, v in value.iteritems() if k != '$tag'])
        if self.__schema_a.tag == tag:
            self.__schema_a.validate(new_val)
        else:
            if isinstance(self.__schema_b, TagSchema):
                if self.__schema_b.tag == tag:
                    self.__schema_b.validate(new_val)
                else:
                    raise JsonSchemaError(u'未知のタグ: %s' % tag)
            elif isinstance(self.__schema_b, UnionSchema):
                self.__schema_b.validate(value)

    def _validate_normal_union(self, value):
        u"""通常の union 型に対するスキーマ適用処理。
        """
        try:
            self.__schema_a.validate(value)
        except JsonSchemaError, e1:
            try:
                self.__schema_b.validate(value)
            except JsonSchemaError, e2:
                raise JsonSchemaError(str(e1) + ' / ' + str(e2))
            else:
                return
        else:
            return

    def convert(self, value):
        try:
            return self.__schema_a.convert(value)
        except JsonSchemaError, e1:
            try:
                return self.__schema_b.convert(value)
            except JsonSchemaError, e2:
                raise JsonSchemaError(u'ユニオン型への変換に失敗しました。' + str(e1) + str(e2))

    def dump(self, depth=0):
        return self.__schema_a.dump(depth) + '|' + self.__schema_b.dump(depth)

class TagSchema(SchemaBase):
    u"""タグ付き union で使われるスキーマ。
    通常のスキーマに tag プロパティが付いたものである。
    """
    def __init__(self, tag, schema):
        self.__schema = schema
        self.__tag = tag

    def create(self, options, underlyingtypes):
        u"""タグ付きスキーマにオプションや型制約が指定がされた場合、
        それは自身の包含しているスキーマへ委譲される。
        """
        return TagSchema(self.__tag, self.__schema.create(options, underlyingtypes))

    @property
    def tag(self):
        return self.__tag

    def validate(self, value):
        self.__schema.validate(value)

    def _convert(self, value):
        return self.__schema.convert(value)

class ScalarSchema(SchemaBase):
    def create(self, options, underlyingtypes):
        if underlyingtypes:
            raise JsonSchemaError(u'スカラー型に underlyingtypes [%s] は適用できません' % str(underlyingtypes))
        return self.__class__(**options)

    def convert(self, value):
        u"""Web からの入力は常にリスト形式となる。
        そのため、1要素のリストは常にスカラー値に変換する。
        """
        if isinstance(value, list) and len(value) == 1:
            return SchemaBase.convert(self, value[0])
        else:
            return SchemaBase.convert(self, value)

class NumberSchema(ScalarSchema):
    def __init__(self, minimum=None, maximum=None, optional=False, is_integer=False):
        self.minimum=minimum
        self.maximum = maximum
        self.optional = optional
        self.is_integer = is_integer

    def create(self, options, underlyingtypes):
        if underlyingtypes:
            raise JsonSchemaError(u'数値型に underlyingtypes [%s] は適用できません' % str(underlyingtypes))
        if self.minimum:
            options['minimum'] = self.minimum
        if self.maximum:
            options['maximum'] = self.maximum
        if self.optional:
            options['optional'] = self.optional
        if self.is_integer:
            options['is_integer'] = self.is_integer
        return NumberSchema(**options)


    def validate(self, value):
        if not self.optional and value == None:
            raise JsonSchemaError(u'null は許可されていません')
        elif self.optional and value == None:
            return
        if self.is_integer and not isinstance(value, int) or self.is_integer and isinstance(value, bool):
            raise JsonSchemaError(u'%s は整数ではありません' % str(value))
        if not isinstance(value, (float, int)):
            raise JsonSchemaError(u'%s は数値ではありません' % str(value))
        if self.minimum != None and self.minimum > value:
            raise JsonSchemaError(u'値は %f 未満でなければなりません' % self.minimum)
        if self.maximum != None and self.maximum < value:
            raise JsonSchemaError(u'値は %f よりも大きく泣ければなりません' % self.maximum)

    def _convert(self, value):
        if self.is_integer:
            return int(value)
        return float(value)

    def dump(self, depth):
        return 'number'

class StringSchema(ScalarSchema):
    def __init__(self, minLength=None, maxLength=None, optional=False, encoding='utf-8'):
        self.minLength = minLength
        self.maxLength = maxLength
        self.optional = optional
        self.encoding = encoding

    def validate(self, value):
        if not self.optional and value == None:
            raise JsonSchemaError(u'null は許可されていません')
        elif self.optional and value == None:
            return
        if not isinstance(value, unicode):
            raise JsonSchemaError(u'値は unicode でなければなりません')
        if self.minLength != None and self.minLength > len(value):
            raise JsonSchemaError(u'文字列長は %d よりも長くなければなりません' % self.minLength)
        if self.maxLength != None and self.maxLength < len(value):
            raise JsonSchemaError(u'文字列長は %d よりも短くなければなりません' % self.maxLength)

    def _convert(self, value):
        if value == None or isinstance(value, unicode):
            return value
        elif isinstance(value, basestring):
            return unicode(value, self.encoding)
        else:
            raise JsonSchemaError(u'文字列のみが変換対象です')

    def dump(self, depth):
        return 'string'

class BinarySchema(ScalarSchema):
    def __init__(self, minLength=None, maxLength=None, optional=False):
        self.minLength = minLength
        self.maxLength = maxLength
        self.optional = optional

    def validate(self, value):
        if not self.optional and value == None:
            raise JsonSchemaError(u'null は許可されていません')
        elif self.optional and value == None:
            return
        if not isinstance(value, str):
            raise JsonSchemaError(u'値はバイト列でなければなりません')
        if self.minLength != None and self.minLength > len(value):
            raise JsonSchemaError(u'データ長は %d よりも大きくなければなりません' % self.minLength)
        if self.maxLength != None and self.maxLength < len(value):
            raise JsonSchemaError(u'データ長は %d よりも小さくなければなりません' % self.maxLength)

    def _convert(self, value):
        if value == None or isinstance(value, str):
            return value
        return str(value)

    def dump(self, depth):
        return 'string'


class BoolSchema(ScalarSchema):
    def __init__(self, optional=False):
        self.optional = optional

    def validate(self, value):
        if not self.optional and value == None:
            raise JsonSchemaError(u'null は許可されていません')
        elif self.optional and value == None:
            return
        if not isinstance(value, bool):
            raise JsonSchemaError(u'値は真偽値でなければなりません')

    def _convert(self, value):
        if isinstance(value, basestring):
            return eval(value.capitalize())
        else:
            return bool(value)

    def dump(self, depth):
        return 'boolean'

class ArraySchema(SchemaBase):
    def __init__(self, repeat=False, optional=False, maxItems=None):
        self.optional = optional
        self.repeat = repeat
        self.schemas = [AnySchema()]
        self.maxItems = maxItems

    def create(self, options, underlyingtypes):
        r = self.__class__(**options)
        if underlyingtypes:
            return r._apply_udt(underlyingtypes)
        return r

    def _apply_udt(self, args):
        if not hasattr(args, '__iter__'):
            args = [args]
        if self.maxItems and len(args) > self.maxItems:
            raise JsonSchemaError(u'要素数は %d 個以内でなければなりません' % (len(self.schemas)))
        obj = self.clone()
        obj.schemas = []
        for s in args:
            obj.schemas.append(s)
        return obj
    
    def validate(self, value):
        if not self.optional and value == None:
            raise JsonSchemaError(u'null は許可されていません')
        elif self.optional and not value:
            return
        if not (isinstance(value, list) or isinstance(value, tuple)):
            raise JsonSchemaError(u'値はリストでなければなりません')
        if self.maxItems and len(value) > self.maxItems:
            raise JsonSchemaError(u'この型には %d 個以内の要素数しか挿入できません' % (len(self.schemas)))
        index = 0
        for v in value:
            if index >= len(self.schemas):
                if not self.repeat:
                    raise JsonSchemaError(u'この型には %d 個以内の要素数しか挿入できません' % (len(self.schemas)))
                elif self.repeat:
                    index = len(self.schemas) - 1
            self.schemas[index].validate(v)
            index += 1

    def _convert(self, value):
        if not (isinstance(value, list) or isinstance(value, tuple)):
            raise JsonSchemaError(u'値はリストでなければなりません')
        def _():
            index = 0
            for v in value:
                if index >= len(self.schemas):
                    if not self.repeat:
                        raise JsonSchemaError(u'この型には %d 個以内の要素数しか挿入できません' % (len(self.schemas)))
                    elif self.repeat:
                        index = len(self.schemas) - 1
                yield self.schemas[index].convert(v)
                index += 1
        return list(_())

    def dump(self, depth):
        r = []
        for s in self.schemas:
            r.append(s.dump(depth))
        if self.repeat:
            r[-1] += '*'
        return ', '.join(r)

class ListSchema(ArraySchema):
    def __init__(self, optional=False, maxItems=None):
        ArraySchema.__init__(self, optional=optional, maxItems=maxItems, repeat=True)

    def _apply_udt(self, arg):
        if len(arg) > 1:
            raise JsonSchemaError(u'指定できる型は一つまでです')
        obj = self.clone()
        return ArraySchema._apply_udt(self, arg)

class TupleSchema(ArraySchema):
    def __init__(self, optional=False, maxItems=None):
        ArraySchema.__init__(self, optional=optional, maxItems=maxItems, repeat=False)

    def _apply_udt(self, args):
        obj = self.clone()
        return ArraySchema._apply_udt(self, args)

class EnumSchema(SchemaBase):
    def __init__(self, optional=False):
        self.optional = optional
        self.enum = []

    def create(self, options, enum):
        e = EnumSchema(self.optional)
        if 'optional' in options:
            e.optional = options['optional']
        e.enum = enum
        return e

    def validate(self, value):
        if not self.optional and not value:
            raise JsonSchemaError(u'null は許可されていません')
        elif self.optional and not value:
            return
        if not value in self.enum:
            raise JsonSchemaError(u'値は次の要素のうちのどれかでなければなりません: %s' % (str(self.enum)))

    def _convert(self, value):
        if isinstance(self.enum[0], StringSchema):
            return StringSchema()._convert(value)
        else:
            return IntegerSchema()._convert(value)

    def dump(self, depth):
        return ', '.join(['"%s"' % s for s in self.enum])

class MultiValueSchema(EnumSchema):
    def validate(self, values):
        for v in values:
            EnumSchema.validate(self, v)

    def _convert(self, values):
        return list((EnumSchema.convert(self, v) for v in values))

class IntersectionSchema(SchemaBase):
    def __init__(self, a, b):
        self.schema_a = a
        self.schema_b = b

    def validate(self, value):
        self.schema_a.validate(value)
        self.schema_b.validate(value)

class ObjectSchema(SchemaBase):
    u"""JSON オブジェクトの妥当性検証を行うクラス。
    """
    def __init__(self, optional=False, wildcard=None, schemas=None):
        self.optional = optional
        if wildcard == None: wildcard = AnySchema()
        self._wildcard = wildcard
        self.schemas = schemas if schemas else {}

    def wildcard():
        def _get(self):
            return self._wildcard

        def _set(self, v):
            return self._wildcard

        return _get, _set
    wildcard = property(*wildcard())

    def create(self, options, underlyingtypes):
        o = ObjectSchema(**options)
        o.schemas = copy.deepcopy(self.schemas)
        o.wildcard = self.wildcard
        return o

    def __and__(self, another):
        return IntersectionSchema(self, another)

    def validate(self, value):
        if not self.optional and value == None:
            raise JsonSchemaError(u'null は許可されていません')
        elif self.optional and value == None:
            return
        if not isinstance(value, dict):
            raise JsonSchemaError(u'値は JSON object (dict) でなければなりません')
        for k, v in value.iteritems():
            if k not in self.schemas:
                if self.wildcard is None:
                    raise JsonSchemaError(u'未知のプロパティ: %s' % k)
                else:
                    self.wildcard.validate(v)
            else:
                try:
                    self.schemas[k].validate(v)
                except Exception, e:
                    raise JsonSchemaError('%s: %s' % (k, str(e)))
        # optional でないメンバーで漏れがないかチェック
        for k, v in self.schemas.iteritems():
            if not v.optional and k not in value:
                raise JsonSchemaError(u'プロパティが足りません: %s' % k)

    def _convert(self, value):
        if not isinstance(value, dict):
            raise JsonSchemaError(u'値は JSON object (dict) でなければなりません')
        result = {}
        for k, v in value.iteritems():
            if k not in self.schemas:
                if self.wildcard is None:
                    raise JsonSchemaError(u'未知のプロパティ: %s' % k)
                else:
                    result[k] = self.wildcard.convert(v)
            else:
                result[k] = self.schemas[k].convert(v)
        return result

    def __iter__(self):
        return self.schemas.iteritems()

    def dump(self, depth=0):
        s = []
        s.append('    ' * depth)
        s.append('{\n')
        depth += 1
        for k, v in self.schemas.iteritems():
            s.append('    ' * depth)
            s.append(k)
            s.append(':')
            s.append(v.dump(depth))
            s.append('\n')
        if self.wildcard:
            s.append('    ' * depth)
            s.append('*')
            s.append(':')
            s.append('any')
            s.append('\n')
        s.append('    ' * (depth - 1))
        s.append('}')
        return ''.join(s)

    def __getitem__(self, key):
        return self.schemas[key]

class AnySchema(SchemaBase):
    u"""あらゆる型を受け付けるスキーマ。
    つまり何もしない。
    """

    def create(self, *args, **kwds):
        return self

    def validate(self, value):
        pass

    def _convert(self, value):
        return value

    def dump(self, depth):
        return 'any'

class NullSchema(SchemaBase):
    u"""値が null 値、つまり None であれば良とするスキーマ。
    """
    def validate(self, value):
        if value is not None:
            raise JsonSchemaError(u'値が null ではありません')

    def create(self, *args, **kwds):
        return self

    def _convert(self, value):
        return value

    def dump(self, depth):
        return 'null'

class VoidSchema(SchemaBase):
    u"""Null と同じ unit だが、値なしを意味するところが違う。
    """

    def create(self, *args, **kwds):
        return self

    def validate(self, value):
        pass

    def _convert(self, value):
        return None

    def dump(self, depth):
        return 'void'

class NeverSchema(SchemaBase):
    u"""そもそもアクセスがあってはいけないスキーマ。
    """
    def validate(self, value):
        raise JsonSchemaError('never')

    def create(self, *args, **kwds):
        return self

    def _convert(self, value):
        raise JsonSchemaError('never')

    def dump(self, depth):
        return 'never'

# 組み込み・デフォルト定義のスキーマ群
schemas = {
    'integer': NumberSchema(is_integer=True),
    'number': NumberSchema(),
    'string': StringSchema(),
    'binary': BinarySchema(),
    'boolean': BoolSchema(),
    'array': ArraySchema(),
    'list': ListSchema(),
    'tuple': TupleSchema(),
    'enum': EnumSchema(),
    'object': ObjectSchema(),
    'any': AnySchema(),
    'null': NullSchema(),
    'void': VoidSchema(),
    'never': NeverSchema(),
    'date': ObjectSchema(schemas={'year': StringSchema(), 'month': StringSchema(), 'day': StringSchema()}),
    'location': ObjectSchema(schemas={'status': StringSchema(), 
                                      'header': ObjectSchema(schemas={'Location':StringSchema()})
                                     }),
    'setcookie': ObjectSchema(schemas={'header': ObjectSchema(schemas={'Set-Cookie': StringSchema()})})
}

