# coding: utf-8
__all__ = ['check_grant', 'simple_checker', 'null_checker', 'AuthoriToken', 'CREATE', 'READ', 'UPDATE', 'DELETE']
import os

CREATE = 0
READ = 1
UPDATE = 2
DELETE = 3

_OP_MAP = {
    CREATE: 'CREATE',
    READ: 'READ',
    UPDATE: 'UPDATE',
    DELETE: 'DELETE',
    }

class AuthoriToken(object):
    u"""認可トークンクラス。
    mafs では認証については関知しないが read/write などの処理の認可を
    その責務として持っている。
    認可トークンのベースクラスは単に OWNER/GUEST の用なトークン値を持つだけであり、
    そのトークンをどのようにしてアクセス制御を行うかについては
    個々の mafs バックエンドで決定することとする。
    """
    def __init__(self, token_value):
        self.__token_value = token_value

    @property
    def value(self):
        return self.__token_value

def check_grant(verb, fmap):
    u"""権限チェック関数。
    以下のようにデコレータとして使うことで、権限チェックを行う。

    @check_grant(CREATE, fmap)
    def createFile(authori_token, path, metadata): ...

    fmap は CRUD 処理に対応した関数の辞書であり、型を書くと

    int: AuthoriToken -> str -> bool

    という物になっている。
    この fmap によりアクセスコントロールの実装を差し替えることができる。
    実際には simple_checker などのラッパー関数を使い、以下のような API とすることを推奨する。

    @simple_checker(CREATE)
    def createFile(authori_token, path, metadata): ...
    """
    def wrapper(f):
        def check(token, path, *args, **kwds):
            checker_func = fmap.get(verb, None)
            if not checker_func:
                raise Exception('Undefined operation: %s, %s' % (path, str(verb)))
            if checker_func(token, path):
                return f(token, path, *args, **kwds)
            else:
                raise Exception('Operation not allowed: %s, %s' % (path, _OP_MAP[verb]))
        return check
    return wrapper

def simple_checker(verb):
    u"""単純な権限チェック関数。
    check_grant に渡す fmap のリファレンス実装と思ってよい。
    この実装では以下の規約に則ったアクセス制御が行われる。

    * admin はあらゆるファイルの CRUD が可能である
    * member は /admin/ 以下のファイルでないファイルを読み取り可能である
      また、 /share/ 以下のファイルであれば作成・書き込み可能である（削除はできない）
    * guest は /public/ と /share/ 以下のファイルのみを読み取り可能である
    """

    admin = 'admin'
    member = 'member'
    guest = 'guest'

    def check_create(token, path):
        v = token.value
        if path.startswith('/public/'): return True
        if v == admin:
            return True
        if v == member and path.startswith('/share/'):
            return True
        else:
            return path.startswith('/share/')

    def check_read(token, path):
        v = token.value
        if v == admin:
            return True
        elif v == member:
            if path.startswith('/admin/'):
                return False
            return True
        elif v == guest:
            return path.startswith('/public/') or path.startswith('/share/')
        else:
            raise Exception('Unknown authorization token')

    def check_delete(token, path):
        v = token.value
        if v == admin:
            return True
        else:
            return False

    def check_update(token, path):
        v = token.value
        if path.startswith('/public/'): return True
        if v == admin:
            return True
        if v == member and path.startswith('/share/'):
            return True
        else:
            return path.startswith('/share/')

    fmap = {
        CREATE: check_create,
        READ: check_read,
        UPDATE: check_update,
        DELETE: check_delete,
    }

    return check_grant(verb, fmap)

def null_checker(verb):
    u"""まったく権限チェックを行わない。
    """
    fmap = {
        CREATE: lambda a, b: True,
        READ: lambda a, b: True,
        UPDATE: lambda a, b: True,
        DELETE: lambda a, b: True,
    }

    return check_grant(verb, fmap)

