# coding: utf-8
u"""mafs バックエンドの初期化モジュール。
mafs では標準関数の実装とパス名や権限チェックは切り離され、
自由に組み合わせて使うことが可能である。
このモジュールではそれらのチェック関数と mafs 実装モジュールを組み合わせ、
新たなモジュールを作成する関数を提供する。
"""
import types
from caty.mafs.authorization import simple_checker, CREATE, READ, UPDATE, DELETE
from caty.mafs.pathname import default_path_checker
from caty.mafs.pathname import path_checker_combinator as pcc
from caty.mafs.pathname import chroot
from caty.mafs.pathname import trimroot
from caty.mafs import fileobject
from caty.mafs import stdfs

__all__ = ['build_mafs_module', 'build_fo_module']

def build_mafs_module(fs=stdfs, root_dir='', grant_checker=simple_checker, path_checker=default_path_checker):
    u"""mafs モジュール初期化関数。
    fs は mafs 標準関数を定義したモジュールその物である。
    権限チェック関数及びパス名のチェック関数を受け取り、
    それらを fs で定義された標準関数と結合させ、
    結合後の関数を公開メンバとして持つオブジェクトを返す。
    また root_dir を指定することで、 mafs がアクセスできるディレクトリツリーのルートを指定できる。
    デフォルトでは stdfs に simple_checker と default_path_checker を適用し、
    ルートディレクトリの指定を行わないモジュールを返す。
    """
    functions = {
        'putFile': pcc(path_checker, root_dir, grant_checker(CREATE), (fs.putFile)),
        'createFile': pcc(path_checker, root_dir, grant_checker(CREATE), (fs.createFile)),
        'createDirectory': pcc(path_checker, root_dir, grant_checker(CREATE), (fs.createDirectory)),
        'readFile': pcc(path_checker, root_dir, grant_checker(READ), (fs.readFile)),
        'delete': pcc(path_checker, root_dir, grant_checker(DELETE), (fs.delete)),
        'deleteFile': pcc(path_checker, root_dir, grant_checker(DELETE), (fs.deleteFile)),
        'deleteDirectory': pcc(path_checker, root_dir, grant_checker(DELETE), (fs.deleteDirectory)),
        'writeFile': pcc(path_checker, root_dir, grant_checker(UPDATE), (fs.writeFile)),
        'getMetadata': pcc(path_checker, root_dir, grant_checker(READ), (fs.getMetadata)),
    }
    # readDirectory は内部的に getMetadata を使用しており、その参照先が元の権限チェックなしのままでは不都合。
    # ややアドホックなやり方ではあるが、権限チェック機構を付加した getMetadata を明示的に渡すことで対処。
    functions['readDirectory'] = pcc(path_checker, root_dir, grant_checker(READ), 
        lambda token, path: fs.readDirectory(token, path, trimroot(root_dir, functions['getMetadata']))
    )
    module = types.ModuleType('mafsimplement')
    module.__dict__.update(functions)
    return module

def build_fo_module(module):
    u"""引数の module をバックエンドとして用いるクラスと関数を新たに構築し、
    それらをメンバーとするモジュールを返す。
    """
    class FileObject(fileobject.FileObject):
        def __init__(self, auth_token, path, mode):
            fileobject.FileObject.__init__(self, auth_token, path, mode, module)

    class DirectoryObject(fileobject.DirectoryObject):
        def __init__(self, auth_token, path):
            fileobject.DirectoryObject.__init__(self, auth_token, path, module)

    def open(auth_token, path, mode='rb'):
        return FileObject(auth_token, path, mode)

    objects = {
        'FileObject': FileObject,
        'DirectoryObject': DirectoryObject,
        'open': open,
    }

    fom = types.ModuleType('mafsfo')
    fom.__dict__.update(objects)
    return fom

