# coding: utf-8
from __future__ import with_statement
import os
from datetime import datetime
from caty.mafs import stdfs
from caty.mafs.metadata import timestamp, caty_dec

class FileObject(object):
    u"""mafs のファイルオブジェクト。
    このクラス自体はファイルシステムへのアクセス機能を持たず、
    個々の mafs 実装で定義された関数へのラッパーオブジェクトとして機能する。
    """
    def __init__(self, auth_token, path, mode, module):
        self.__path = path
        self.__auth_token = auth_token
        self.__module = module
        self.__buffer = []
        self.__mode = mode
        self.__closed = False
        self.__read = False
        if mode == 'wb':
            self._create()

    @property
    def path(self):
        return self.__path

    @property
    def auth_token(self):
        return self.__auth_token

    @property
    def basename(self):
        return self.path.rsplit('/', 1)[-1]

    @property
    def content_length(self):
        return self.__module.getMetadata(self.auth_token, self.path).contentLength

    @property
    def last_modified(self):
        return self.__module.getMetadata(self.auth_token, self.path).lastModified

    @property
    def content_type(self):
        return unicode(self.__module.getMetadata(self.auth_token, self.path).contentType)

    @property
    def is_text(self):
        return self.__module.getMetadata(self.auth_token, self.path).isText

    @property
    def encoding(self):
        return unicode(self.__module.getMetadata(self.auth_token, self.path).textEncoding)

    @property
    def exists(self):
        try:
            self.last_modified
            return True
        except:
            return False

    @property
    def is_executable(self):
        return self.__module.getMetadata(self.auth_token, self.path).executable

    @property
    def is_hidden(self):
        return self.__module.getMetadata(self.auth_token, self.path).hidden

    def mode():
        def _get(self):
            return self.__mode
        def _set(self, value):
            if not value in  ('rb', 'wb'):
                raise Exception('Invalid mode: %s' % value)
            self.__mode = value
        return _get, _set
    mode = property(*mode())

    @property
    def closed(self):
        return self.__closed

    def closed_check(f):
        def _(self, *args, **kwds):
            if self.closed:
                raise Exception('Already closed: %s' % self.path)
            else:
                return f(self, *args, **kwds)
        return _

    @closed_check
    def read(self, remove=True, encode=True):
        u"""ファイルの内容をすべて読み出す。
        読み取り権限を持たない場合、例外を送出する。
        remove が True の場合、 caty-encoding 宣言を削除した上で読み出す（デフォルトの動作）。
        encode 及び is_text が True の場合、ファイルの内容を unicode にして返す（デフォルトの動作）。
        """
        data = self.__module.readFile(self.auth_token, self.path)
        if self.__read: return ''
        if remove:
            data = caty_dec.sub('', data)
        self.__read = True
        if self.is_text and self.encoding:
            return unicode(data, self.encoding)
        else:
            return data

    def _create(self):
        u"""空のファイルを作成する。
        """
        self.__module.createFile(self.auth_token, self.path)

    @closed_check
    def write(self, content):
        u"""ファイルへの書き込みを行う。
        ファイルが存在しない場合はエラーとなる。
        一連のファイルへの書き込みはアトミックなものであり、
        書き込みが途中まで成功したがその後失敗などという状態にはならない。
        """
        self.__buffer.append(content)
    
    @closed_check
    def close(self):
        u"""ファイルを閉じる。
        一旦閉じられたファイルオブジェクトは再利用できない。
        書き込みモードであり尚且つバッファにデータがある場合、それをファイルに書き込む。
        """
        if self.mode == 'wb' and self.__buffer:
            self.__module.writeFile(self.auth_token, self.path, ''.join(self.__buffer))
        self.__closed = True

    @closed_check
    def delete(self, dummy=False):
        r"""ファイルを削除する。
        dummy 引数は DirectoryObject とのインターフェースを合致させるためのものである。
        """
        self.__module.deleteFile(self.auth_token, self.path)

    @closed_check
    def is_modifed_since(self, ts):
        u"""引数の時刻とファイルのタイムスタンプを比べ、
        ファイルの方が新しければ True を返す。
        引数の型は float もしくは datetime オブジェクトのみを受け付ける。
        """
        if isinstance(ts, float):
            ts = timestamp(ts)
        elif not isinstance(ts, datetime):
            raise TypeError('Type of argument must be float or datetime')
        return self.last_modified > ts

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not self.closed:
            self.close()
        if not exc_type and not exc_value and not traceback:
            return True
        else:
            return False

    def __del__(self):
        if not self.closed:
            self.close()

    @property
    def is_dir(self):
        return False

class DirectoryObject(object):
    u"""mafs におけるディレクトリ。
    """
    def __init__(self, auth_token, path, module):
        self.__path = path
        self.__auth_token = auth_token
        self.__module = module

    @property
    def path(self):
        return self.__path

    @property
    def auth_token(self):
        return self.__auth_token

    @property
    def last_modified(self):
        return self.__module.getMetadata(self.auth_token, self.path).lastModified

    @property
    def is_hidden(self):
        return self.__module.getMetadata(self.auth_token, self.path).hidden

    @property
    def exists(self):
        try:
            self.last_modified
            return True
        except:
            return False

    def create(self):
        r"""ディレクトリを作成する。
        """
        self.__module.createDirectory(self.auth_token, self.path)

    def read(self, recursive=False):
        r"""path 直下のディレクトリエントリを FileObject または DirectoryObject のジェネレータとして返す。
        recursive = True の時、再帰的にディレクトリをたどる。
        """
        for e in self.__module.readDirectory(self.auth_token, self.path):
            if e.is_dir:
                if not recursive:
                    yield DirectoryObject(self.auth_token, os.path.join(self.path, e.name), self.__module)
                else:
                    for se in DirectoryObject(self.auth_token, os.path.join(self.path, e.name), self.__module).read(recursive):
                        yield se
            else:
                yield FileObject(self.auth_token, os.path.join(self.path, e.name), 'rb', self.__module)

    def delete(self, recursive=False):
        u"""ディレクトリを削除する。
        recursive が True の場合、再帰的にディレクトリを削除する。
        """
        if recursive:
            for e in self.read():
                e.delete(recursive)
            self.__module.deleteDirectory(self.auth_token, self.path)
        else:
            self.__module.deleteDirectory(self.auth_token, self.path)

    @property
    def is_dir(self):
        return True

