#coding: utf-8
u"""mafs メタデータモジュール。
mafs バックエンドの実装者以外はこのモジュールを直接使う必要はない。
"""

import os
import time
import datetime

TEXT_TYPES = [
    'text/plain',
    'text/html',
    'text/xml',
    'text/css',
    'application/xml',
    'application/rss+xml',
    'application/json',
]

MIME_TYPES = {
    '.htm': 'text/html',
    '.css': 'text/css',
    '.html': 'text/html',
    '.xhtml': 'text/html',
    '.xml': 'text/xml',
    '.txt': 'text/plain',
    '.rdf': 'application/rss+xml',
    '.jpg': 'image/jpeg',
    '.jpeg': 'image/jpeg',
    '.png': 'image/png',
    '.json': 'application/json',
}

import re
caty_dec = re.compile(r'<\?caty-encoding +encoding *= *"(.+?)" *\?>' + '\n?')
xml_dec = re.compile(r'<\?xml +version *= *"[0-9.]+" *\?>')
xml_dec_with_enc = re.compile(r'<\?xml +version *= *"[0-9.]+" +encoding *= *"(.+?)" *\?>')

def timestamp(sec):
    return datetime.datetime(*(time.localtime(sec)[:-3]))

def mime_type(ext):
    if not ext: return 'application/octet-stream'
    return MIME_TYPES.get(ext, 'application/octet-stream')

class FileMetadata(object):
    def __init__(self, 
            contentLength, 
            lastModified, 
            contentType,
            executable = False,
            hidden = False,
            textEncoding = 'utf-8',
            readOnly = False,
            created = 0):
        self.__contentLength = contentLength
        self.__lastModified = timestamp(lastModified)
        self.__contentType = contentType
        self.__textEncoding = textEncoding
        self.__executable = executable
        self.__hidden = hidden
        self.__readOnly = readOnly
        if created:
            self.__created = timestamp(created)
        else:
            self.__created = None

    @property
    def contentLength(self):
        return self.__contentLength

    @property
    def lastModified(self):
        return self.__lastModified

    @property
    def contentType(self):
        return self.__contentType
        
    @property
    def isText(self):
        return self.contentType in TEXT_TYPES

    def textEncoding():
        def _get(self):
            if not self.isText: return ''
            return self.__textEncoding
        def _set(self, value):
            if not self.isText: return
            self.__textEncoding = value
        return _get, _set
    textEncoding = property(*textEncoding())
    
    @property
    def executable(self):
        return self.__executable

    @property
    def hidden(self):
        return self.__hidden

    @property
    def readOnly(self):
        return self.__readOnly

    @property
    def created(self):
        return self.__created

class DirectoryMetadata(object):
    def __init__(self, lastModified, hidden=False, readOnly=False, created=0):
        self.__lastModified = timestamp(lastModified)
        self.__hidden = hidden
        self.__readOnly = readOnly
        if created:
            self.__created = timestamp(created)
        else:
            self.__created = None

    @property
    def lastModified(self):
        return self.__lastModified

    @property
    def readOnly(self):
        return self.__readOnly

    @property
    def hidden(self):
        return self.__hidden
        
    @property
    def created(self):
        return self.__created

class DirectoryEntry(object):
    def __init__(self, path, metadata):
        self.__name = os.path.basename(path)
        self.__metadata = metadata

    @property
    def name(self):
        return self.__name

    @property
    def metadata(self):
        return self.__metadata

    @property
    def is_dir(self):
        return isinstance(self.metadata, DirectoryMetadata)

