# coding: utf-8
from caty.shell.template import Template
import caty.jsontools as json
from caty.jsontools.model import tagged
from caty.shell.script.cmdbase import Builtin
import sys
import traceback
import codecs
from optparse import Values
from StringIO import StringIO
from threading import RLock
__all__ = ['Print', 'Today', 'Login', 'Loggedin', 'Adduser', 'Addloggedinfo']

class Print(Builtin):
    u"""組み込みの print コマンド。
    """
    command_decl = u"""
    command print {"raw": boolean, "resolve": boolean, "verbose": boolean} [string] :: 
        any -> object {
            "header": object {
                *: any
            },
            "body": string | binary
        }
        updates mafs
        reads stream
        refers python:caty.shell.script.builtin.Print;
    """

    def set_facility(self, facilities):
        from caty.mafs import filesystem
        from caty.mafs.authorization import AuthoriToken
        Builtin.set_facility(self, facilities)
        self.mafs = filesystem.bind_token(AuthoriToken('admin'))

    def dup(self):
        p = self.__class__(self.opt_and_args)
        p.mafs = self.mafs
        p.env = self.env
        p.config = self.config
        p.var = self.var
        p.stream = self.stream
        return p

    def _load_context(self, path):
        u"""path に対応した caty スクリプトファイルを実行する。
        """
        builder = self.builder
        assoc_path =[path+ '.caty', path + '.icaty']
        for p in assoc_path:
            if self.mafs.FileObject(p, 'rb').exists:
                content = self.mafs.open(p).read()
                cmd = builder.build(content)
                if self.input != None:
                    cmd.input = cmd.convert(self.input)
                r = cmd()
                return r, cmd.output
        return None, None

    def subtemplate(self, path):
        opts = Values({
            'resolve': True,
            'raw': None,
            'verbose': False}
        )
        sub = self.dup()
        sub.mafs = self.mafs
        sub.config = self.config
        sub.builder = self.builder
        sub.execute(opts, path)
        return sub.output['body']
    
    def execute(self, opts, path):
        assoc = False
        outfile = codecs.getwriter(self.stream.encoding)(StringIO())
        if opts.raw:
            fo = self.mafs.open(path)
            body = fo.read(remove=False)
            length = len(body)
            self.output = {
                'header': {
                    'content-type': fo.content_type, 
                    'content-length': length
                }, 
                'body': body}
            return 
        if opts.resolve:
            assoc = True
        try:
            template = Template(path, assoc, self.mafs)
            template.set_include_callback(lambda p: self.subtemplate(p))
            if assoc:
                r, o = self._load_context(path)
                if r != None:
                    self.output = o
                    return r
            elif self.input == None:
                self.input = self.stream.read()
            fo = self.mafs.open(path)
            template.context = self.input
            template.write(outfile)
            outfile.seek(0)
            body = unicode(outfile.read(), fo.encoding)
            self.output = {
                    'header': {
                        'content-type': fo.content_type + (';charset=%s' % (fo.encoding)) if fo.encoding else '',
                        'content-length': len(body.encode(fo.encoding)),
                     },
                    'body': body}
        except Exception, e:
            print traceback.format_exc()
            raise
        return 0

from caty.shell.user import login
from caty.shell.session import SessionStorage
from caty.mafs.authorization import AuthoriToken
from caty.jsontools import schema
from Cookie import Morsel
class Login(Builtin):
    u"""ログイン処理。
    """
    command_decl = u"""command login
                        :: object  {"username": string, 
                                   "password": string, 
                                   "redirect": string?} -> null
                               breaks @success location | @fail string
                        refers python:caty.shell.script.builtin.Login;
    """

    def execute(self, opts):
        username = self.input['username']
        password = self.input['password']
        succ, grant = login(username, password)
        if succ:
            session = SessionStorage.create()
            self.env = session
            self.env.put_privilege(self, 'loggedin', schema.registrar.get_schema('boolean'), True)
            self.env.put_privilege(self, 'grant', schema.registrar.get_schema('string'), grant)
            if 'redirect' in self.input:
                redirect = self.input['redirect']
            else:
                redirect = '/'
            self.exit(tagged(
                'success',
                {
                    'header': {
                        "Location": unicode(self.config['server_url'] + redirect),
                        "Set-Cookie": unicode(self._mk_cookie(session.key)),
                    },
                    'status': u'302 Found'
                }
            ))
        else:
            self.eixt(tagged(
                'fail',
                unicode(self.config['login_path'])
            ))
        return 0

    def _mk_cookie(self, sessionid):
        m = Morsel()
        m.set('sessionid', sessionid, sessionid)
        m['expires'] = self.config['session_timeout']
        m['path'] = '/'
        return m.OutputString()


class Loggedin(Builtin):
    u"""ログイン状態のチェック処理。
    ログイン状態であれば入力データをそのまま出力に設定して残りのコマンドに引き渡し、
    そうでなければログインページにリダイレクトする。
    リダイレクト先のパスは引数で上書き可能とする。
    """
    command_decl = u"""command loggedin {} [] :: any -> any
                                                 breaks location
                                        {} [string] :: any -> any
                                                 breaks location
                                        refers python:caty.shell.script.builtin.Loggedin;
    """

    def execute(self, opts, redirect_path=''):
        if self.env.exists('loggedin') and self.env.get('loggedin'):
            self.output = self.input
        else:
            if not redirect_path:
                redirect_path = self.config['login_path']
            self.exit({
                'header': {
                    'Location': self.config['server_url'] + redirect_path
                },
                'status': '302 Found'
            })

class Addloggedinfo(Builtin):
    u"""ログイン情報の追加処理。
    任意の object 型に対して loogedin という名前でログインしているか否かの情報を加える。
    """
    command_decl = u"""command addloggedinfo {} [] :: object -> object
        refers python:caty.shell.script.builtin.Addloggedinfo;"""

    def execute(self, opts):
        i = self.input
        if self.env.exists('loggedin') and self.env.get('loggedin'):
            logged = True
        else:
            logged = False
        i.update({'loggedin': logged})
        self.output = i
        return 0

from caty.shell.user import add_user
class Adduser(Builtin):
    u"""新規ユーザの追加。
    このコマンドは admin の AuthoriToken を持っていない限り実行できない。
    追加するユーザ情報は引数として渡すことも入力オブジェクトとして渡すこともできる。
    これはコマンドラインの使い勝手と Caty スクリプトとしての使い勝手の両方に備えるためである。
    """
    command_decl = u"""command adduser {} [] :: object {"username":string, "grant":string, "password": string} -> boolean
                                                updates mafs
                                       {} [string, string, string] :: null -> boolean
                                                updates mafs
                                       refers python:caty.shell.script.builtin.Adduser;
    """

    def execute(self, opts, username=None, grant=None, password=None):
        if self.input != None:
            username = self.input['username']
            grant = self.input['grant']
            password = self.input['password']
        if None in (username, grant, password):
            self.output = False
            return -1
        add_user(username, grant, password)
        self.output = True
        return 0


import datetime
class Today(Builtin):
    command_decl = u"""command today {} [] :: null -> date
                                       refers python:caty.shell.script.builtin.Today;
    """

    def execute(self, opts):
        n = datetime.datetime.now()
        y = u'%04d' % n.year
        m = u'%02d' % n.month
        d = u'%02d' % n.day
        self.output = {'year': y, 'month': m, 'day': d}
        return 0
"""
class Cd(Builtin):
    def execute(self):
        pass
"""
