# coding:utf-8
from __future__ import with_statement
import caty.jsontools as json
from optparse import OptionParser
from itertools import *
from functools import partial
import types

__all__ = ['Command', 'Builtin', 'PipelineInterruption', 'PipelineErrorExit', 'compile_builtin']

class PipelineErrorExit(Exception):
    def __init__(self, o):
        self.json_obj = o

class PipelineInterruption(Exception):
    def __init__(self, o):
        self.json_obj = o

class JsonWrapper(object):
    def __init__(self, profile, mode):
        self.__profile = profile
        self.__value = None
        self.mode = mode

    def value():
        def _get(self):
            return self.__value
        def _set(self, value):
            value = self.convert(value)
            self.__value = value
        return _get, _set
    value = property(*value())

    def validate(self):
        schema = self.__profile if self.mode == 'input' else self.__profile.out_schema
        schema.validate(self.__value)

    def convert(self, value):
        schema = self.__profile if self.mode == 'input' else self.__profile.out_schema
        return schema.convert(value)

def execption_hook(f):
    u"""コマンドでのエラー発生時にそのコマンド名を出力し、
    そのまま例外をスローする。
    """
    def wrapped(self, *args, **kwds):
        try:
            return f(self, *args, **kwds)
        except Exception, e:
            if isinstance(e, PipelineInterruption) or isinstance(e, PipelineErrorExit):
                raise
            print "Error:", self
            raise
    return wrapped

class Command(object):
    u"""コマンド基底クラス。
    すべてのコマンドは入力チャンネルと出力チャンネルを持ち、
    関数のように呼び出し可能であるとする。
    このクラスのサブクラスは profile と profile_container の二つのプロパティを持つ。
    このプロパティの実体は caty.jsontools.schema.commandprofile で定義された
    コマンド型宣言オブジェクトであり、コマンドの入出力の型を定義する。

    これはシステムの初期化時に /schema/ 以下の .casm ファイルで定義された
    コマンドの型指定からインジェクトされる。
    そのため、コマンドクラスの実装に際しては必ずスキーマを書かなければならない。

    mafs, env プロパティはそれぞれ mafs モジュールの各関数/クラスへの参照と
    環境変数へのアクセスオブジェクトであり、これらは常にインジェクトされる。
    他のプロパティは入出力をのぞき、基本的にコマンド型宣言の reads, updates, uses 句に
    基づいてインジェクトされる。型宣言に記述の無い Caty ファシリティにはアクセスできない。
    """

    def __init__(self, opt_and_args):
        self.opt_and_args = opt_and_args
        opts, args, profile = self.profile_container.determine_profile(opt_and_args)
        self.opts = opts
        self.args = args
        self.profile = profile
        self.__input = JsonWrapper(self.profile, 'input')
        self.__output = JsonWrapper(self.profile, 'output')
        self._usage = ''

    def set_facility(self, facilities):
        for f in self.profile.facilities:
            setattr(self, f, facilities[f])
        self.mafs = facilities['mafs']
        self.env = facilities['env']
        self.config = facilities['config']
        self.var = facilities['var']

    def dup(self):
        raise NotImplementedError

    @property
    def in_schema(self):
        return self.profile.in_schema

    @property
    def out_schema(self):
        return self.profile.out_schema

    @execption_hook
    def convert(self, value):
        return self.__input.convert(value)

    def input():
        def _get(self):
            return self.__input.value
        @execption_hook
        def _set(self, value):
            self.__input.value = value
        return _get, _set
    input = property(*input())

    def output():
        def _get(self):
            return self.__output.value
        @execption_hook
        def _set(self, value):
            self.__output.value = value
        return _get, _set
    output = property(*output())

    def options():
        def _get(self):
            return self.__options
        def _set(self, value):
            self.__options = value
        return _get, _set
    options = property(*options())

    @property
    def usage(self):
        return self._usage

    def post_type_check(f):
        def _(self, *args, **kwds):
            r = f(self, *args, **kwds)
            if r == 0:
                self.__output.validate()
            return r
        return _

    @post_type_check
    @execption_hook
    def __call__(self):
        if self.args:
            return self.execute(self.opts, *self.args)
        else:
            return self.execute(self.opts)

    def execute(self, *args, **kwds):
        raise NotImplementedError

    def error(self, json_obj):
        u"""コマンドをエラーとして終了させる。
        このメソッドが呼ばれた後はエラーハンドラーに処理が委譲され、
        後のパイプラインは一切処理されなくなる。
        """
        raise PipelineErrorExit(json_obj)

    def exit(self, json_obj):
        u"""コマンドを json_obj を出力にして終了する。
        """
        raise PipelineInterruption(json_obj)

class Builtin(Command):
    u"""組み込みコマンドのベースクラス。
    組み込みコマンドはコマンド宣言をファイルではなく
    プロパティとして直に記述する。
    この時は文字列としてスキーマを記述し、それをスキーマオブジェクトに変換、
    クラスの属性として追加という流れになる。

    このクラス自体は特に何もせず、マーカーとしてのみ使う。
    """

def compile_builtin(module, registrar):
    u"""ビルトインのコマンドのプロパティで定義されたコマンド宣言をコンパイルする。
    ビルトインのコマンドはすべて以下のプロパティを持っていなければならない。
    
    * command_decl

    このプロパティにはコマンド宣言を直に記述する。
    """
    for item in filter(lambda a: isinstance(a, types.TypeType), module.__dict__.values()):
        if issubclass(item, Builtin) and item != Builtin:
            schema_string = u"%s" % (item.command_decl)
            registrar.compile(schema_string)

