# coding: utf-8
u"""caty シェルのコマンドのパージングを行う。
print --out=foo.html template.html のような文字列を
['print', '--out=foo.html', 'template.html'] のように分解し、
またパイプによるコマンドの入出力の結合を行う。
"""
from StringIO import StringIO
from caty.shell.script.node import *
from caty.shell.script.exception import ScriptError
from caty.shell.script.builder import combine, CommandCombinator, Command
import copy
from caty.shell.script.CatyScriptLexer import (Lexer,
                                               LBRACE,
                                               RBRACE,
                                               STRING,
                                               QSTRING,
                                               OPTION,
                                               LBRACKET,
                                               RBRACKET,
                                               SUBSTITUTION,
                                               LSUBSTITUTION,
                                               PIPE,
                                               COMMA,
                                               DELIMTAGCASE,
                                               NUMBER,
                                               LPAREN,
                                               RPAREN,
                                               LF,
                                               CASE)

class PeekableIterator(object):
    def __init__(self, iterable):
        self.iterble = list(iterable)
        self.cursor = 0
        self.savepoint = 0
        self.length = len(self.iterble)

    def __iter__(self):
        return self

    def next(self):
        if self.length <= self.cursor:
            raise StopIteration()
        r = self.iterble[self.cursor]
        self.cursor += 1
        return r

    def set_cursor(self):
        self.savepoint = self.cursor

    def restore_cursor(self):
        self.cursor = self.savepoint

class State(object):
    u"""状態機械
    """
    def __init__(self, prev_state=None, cmd_builder=None, parent=None, lexer=None):
        self.command = None
        self.parent = parent
        self.current_child = None
        self.lexer = lexer
        if prev_state:
            self.prev_state = prev_state
            self.tokens = []
            self.chunks = []
            self.builder = prev_state.builder
            self.is_terminated = False
        else:
            self.prev_state = prev_state
            self.tokens = []
            self.chunks = []
            self.is_terminated = False
            self.builder = cmd_builder

    def create_chunk(self):
        self.chunks.append(self.builder.build(copy.deepcopy(self.tokens)))
        self.tokens[:] = []

    def back(self):
        if self.current_child:
            return self.current_child.back()
        else:
            return self

    def next(self, token):
        ret = self._next(token.getType(), token.getText())
        if not ret:
            if self.parent:
                return self.parent.next(token)
            raise ScriptError('Syntax Error')
        else:
            return ret

    def peek(self, n):
        self.lexer.set_cursor()
        i = n
        while i > 0:
            try:
                v = self.lexer.next().getType()
            except StopIteration, e:
                self.lexer.restore_cursor()
                return v
            if v != LF:
                i -= 1
        self.lexer.restore_cursor()
        return v

    def _next(self, t, s):
        raise NotImplemantedError

class InitialState(State):
    def _next(self, t, s):
        if t == LBRACE:
            if self.peek(2) in (CASE, DELIMTAGCASE):
                self.current_child = DispatchState(self, parent=self, lexer=self.lexer)
                return self.current_child._next(t, s)
            else:
                self.current_child = EchoState(self, parent=self, lexer=self.lexer)
                return self.current_child._next(t, s)
        elif t == STRING:
            if s == 'switch':
                self.current_child = SwitchState(self, parent=self, lexer=self.lexer)
                return self.current_child
            elif s == 'true' or s == 'false':
                self.current_child = EchoState(self, parent=self, lexer=self.lexer)
                return self.current_child._next(t, s.capitalize())
            else:
                ret = CommandState(self, parent=self, lexer=self.lexer)
                ret.tokens.append(s)
                self.current_child = ret
                return self.current_child
        elif t in (QSTRING, NUMBER, LBRACKET):
            self.current_child = EchoState(self, parent=self, lexer=self.lexer)
            return self.current_child._next(t, s)
        elif t == LF:
            if self.current_child and self.current_child.is_terminated:
                self.build()
                self.is_terminated = True
                return self if not self.parent else self.parent
            else:
                return self.back()
        elif t == PIPE:
            if not self.current_child:
                return
            self.chunks.append(self.current_child.command)
            self.current_child = None
            return self

    def build(self):
        if not self.command:
            if self.current_child:
                self.chunks.append(self.current_child.command)
            self.command = combine(self.chunks)

class EchoState(State):
    def _next(self, t, s):
        if t == QSTRING or t == STRING or t == NUMBER:
            if not self.command:
                cmd = ScalarBuilder(None)
                cmd.set_value(eval(s))
                self.command = cmd
                self.is_terminated = True
                return self
            else:
                raise ScriptError('Syntax Error.')
        elif t == LBRACE:
            # ブレースが出現したら ObjectState に処理をディスパッチ
            self.current_child = ObjectState(self, parent=self, lexer=self.lexer)
            return self.current_child
        elif t == LBRACKET:
            self.current_child = ListState(self, parent=self, lexer=self.lexer)
            return self.current_child

class CommandState(State):
    def build(self):
        if self.tokens: self.create_chunk()
        self.command = combine(self.chunks)

    def _next(self, t, s):
        if t == STRING or t == QSTRING or t == NUMBER or t == OPTION:
            self.tokens.append(s)
            return self
        else:
            self.create_chunk()
            self.build()
            self.is_terminated = True
            return self.parent._next(t, s)

class ObjectState(State):
    def __init__(self, *args, **kwds):
        State.__init__(self, *args, **kwds)
        self.nodes = []
        self.__lvalue = None
        self.current_child = None

    def lvalue():
        def _get(self):
            return self.__lvalue
        def _set(self, value):
            self.__lvalue = value
        return _get, _set
    lvalue = property(*lvalue())

    def rvalue():
        def _get(self):
            u"""子ノードからの代入操作にフックを仕掛け、ノードツリーを構築する。
            """
            return None
        def _set(self, value):
            if isinstance(value, Command):
                node = CommandNode(self.lvalue, value)
            elif isinstance(self.lvalue, Command):
                node = CommandNode(value, self.lvalue)
            else:
                node = ConstNode(self.lvalue, value)
            self.nodes.append(node)
        return _get, _set
    rvalue = property(*rvalue())

    def _next(self, t, s):
        if t == QSTRING:
            self.lvalue = eval(s)
            self.current_child = SubstitutionState(self, parent=self, lexer=self.lexer)
            return self.current_child
        elif t == STRING:
            self.tokens.append(s)
            self.current_child = CommandState(self, parent=SubstitutionState(self, parent=self, lexer=self.lexer))
            return self.current_child
        elif t == COMMA: # 子ノードからの戻り
            if not self.current_child:
                return
            self.current_child = None
            return self
        elif t == RBRACE:
            self.build()
            self.parent.command = self.command
            self.parent.is_terminated = True
            return self.parent

    def build(self):
        obj = ObjectBuilder(None)
        for n in self.nodes:
            obj.add_node(n)
        self.command = obj

class ListState(State):
    def __init__(self, *args, **kwds):
        State.__init__(self, *args, **kwds)
        self.nodes = []

    def _next(self, t, s):
        if t in (QSTRING, NUMBER, STRING, LBRACKET, LBRACE):
            if t == STRING and not s in ('true', 'false'):
                return
            self.current_child = InitialState(self, parent=self, lexer=self.lexer)
            return self.current_child._next(t, s)
        elif t == COMMA:
            if self.current_child is not None:
                self.current_child.build()
                self.nodes.append(self.current_child.command)
                self.current_child = None
                return self
        elif t == RBRACKET:
            if self.current_child is not None:
                self.current_child.build()
                self.nodes.append(self.current_child.command)
                self.current_child = None
            self.build()
            self.parent.command = self.command
            self.parent.is_terminated = True
            return self.parent

    def build(self):
        l = ListBuilder(None)
        for n in self.nodes:
            l.add_node(n)
        self.command = l

class SubstitutionState(State):
    def _next(self, t, s):
        if t == SUBSTITUTION:
            # ':' に対応。右辺値はコマンドか定数リテラル
            self.current_child = InitialState(self, parent=self, lexer=self.lexer)
            return self.current_child
        elif t == LSUBSTITUTION:
            # '>:' に対応。この時点で parent.current_child は Command になっている
            # '>:' の場合は左辺値がコマンドになるので、 parent.lvalue にコマンドを設定
            self.parent.current_child.build()
            self.parent.lvalue = self.parent.current_child.command
            self.current_child = NameState(self, parent=self, lexer=self.lexer)
            return self.current_child
        elif t == COMMA or t == RBRACE:
            # , もしくは } が出現したら、左右両辺の値が確定する。
            # なんともアレだが、 self.current_child.value/command を parent.rvalue に突っ込む
            # TODO: echo の入れ子への対応が必要かも
            if isinstance(self.current_child, InitialState):
                self.current_child.build()
                self.parent.rvalue = self.current_child.command
            else:
                self.parent.rvalue = self.current_child.value
            return self.parent._next(t, s)
        
class ConstState(State):
    def __init__(self, *args, **kwds):
        State.__init__(self, *args, **kwds)
        self.value = None

    def _next(self, t, s):
        if t == QSTRING or t == NUMBER or t == STRING:
            if not self.value:
                self.value = eval(s)
                return self

class NameState(State):
    def __init__(self, *args, **kwds):
        State.__init__(self, *args, **kwds)
        self.value = None

    def _next(self, t, s):
        if t == QSTRING:
            if not self.value:
                self.value = eval(s)
                return self

class DispatchState(State):
    def _next(self, t, s):
        if t == LBRACE:
            self.current_child = DispatchBodyState(self, parent=self, lexer=self.lexer)
            return self.current_child


class DispatchBodyState(State):
    def __init__(self, *args, **kwds):
        State.__init__(self, *args, **kwds)
        self.nodes = []
        self.__lvalue = None
        self.current_child = None

    def lvalue():
        def _get(self):
            return self.__lvalue
        def _set(self, value):
            self.__lvalue = value
        return _get, _set
    lvalue = property(*lvalue())

    def rvalue():
        def _get(self):
            u"""子ノードからの代入操作にフックを仕掛け、ノードツリーを構築する。
            """
            return None
        def _set(self, value):
            node = Case(self.lvalue, value)
            self.nodes.append(node)
        return _get, _set
    rvalue = property(*rvalue())

    def _next(self, t, s):
        if t == QSTRING and not self.lvalue:
            self.lvalue = eval(s)
            self.current_child = CaseState(self, parent=self, lexer=self.lexer)
            return self.current_child
        elif t == COMMA: # 子ノードからの戻り
            if not self.current_child:
                return
            self.current_child = None
            self.lvalue = None
            return self
        elif t == RBRACE:
            self.build()
            self.parent.command = self.command
            self.parent.is_terminated = True
            return self.parent

    def build(self):
        obj = Dispatch()
        for n in self.nodes:
            obj.add_case(n)
        self.command = obj

class CaseState(State):
    def _next(self, t, s):
        if t == CASE:
            self.current_child = InitialState(self, parent=self, lexer=self.lexer)
            return self.current_child
        elif t == COMMA or t == RBRACE:
            self.current_child.build()
            self.parent.rvalue = self.current_child.command
            return self.parent._next(t, s)

def parse_command(string, state, builder):
    queue = []
    s = StringIO(string+'\n') # 必ず改行文字を付ける
    lexer = PeekableIterator(Lexer(s))
    state = state if state else InitialState(cmd_builder=builder, lexer=lexer)
    for t in lexer:
        state = state.next(t)
    return state


