#coding: utf-8
u"""Caty スクリプトノードモジュール。
Caty では switch, dispatch, object 生成などはすべてコマンドとして実装する。
"""

from caty.jsontools.path import build_query
from caty.jsontools.model import TaggedValue, tagged
from caty.jsontools import jstypes
from caty.shell.script.exception import ScriptError
from caty.shell.script.cmdbase import Command, Builtin
import types

class ScalarBuilder(Builtin):
    command_decl = u"""command scalar-builder :: void -> any
                        refers python:caty.shell.script.node.ScalarBuilder;
    """

    def __init__(self):
        Builtin.__init__(self, None)

    def set_value(self, value):
        from caty.shell.system import ENCODING
        self.value = value if not isinstance(value, basestring) else unicode(value, ENCODING)

    def execute(self, opts):
        self.output = self.value

class ListBuilder(Builtin):
    command_decl = u"""command list-builder :: any -> any
                        refers python:caty.shell.script.node.ListBuilder;
    """

    def __init__(self):
        Builtin.__init__(self, None)
        self.values= []

    def set_values(self, values):
        for v in values:
            if isinstance(v, Command):
                self.values.append(v)
            else:
                s = ScalarBuilder()
                s.set_value(v)

    def set_facility(self, facilities):
        for v in self.values:
            v.set_facility(facilities)

    def execute(self, opts):
        r = []
        for v in self.values:
            if self.input != None:
                v.input = self.input
            v()
            r.append(v.output)
        self.output = r


class ObjectBuilder(Builtin):
    command_decl = u"""command object-builder :: any -> any
                        refers python:caty.shell.script.node.ObjectBuilder;
    """

    def __init__(self):
        Builtin.__init__(self, None)
        self.__nodes = []

    def add_node(self, node):
        self.__nodes.append(node)

    def set_facility(self, facilities):
        for n in self.__nodes:
            n.set_facility(facilities)

    def execute(self, opts):
        obj = {}
        for node in self.__nodes:
            if self.input != None:
                node.input = self.input
            node()
            obj[node.name] = node.output
        if '$tag' in obj:
            if '$val' in obj:
                obj = tagged(obj['$tag'], obj['$val'])
            else:
                t = obj['$tag']
                del obj['$tag']
                obj = tagged(t, obj)
        self.output = obj


class ConstNode(object):
    def __init__(self, name, value):
        from caty.shell.system import ENCODING
        self.name = name
        self.output = value if not isinstance(value, basestring) else unicode(value, ENCODING)

    def __call__(self):
        return 

class CommandNode(object):
    def __init__(self, name, cmd):
        self.name = name
        self.cmd = cmd
        self.output = None
        self.input = None

    def set_facility(self, facilities):
        self.cmd.set_facility(facilities)

    def __call__(self):
        if self.input != None:
            self.cmd.input = self.input
        self.cmd()
        self.output = self.cmd.output

class Dispatch(Builtin):
    command_decl = u"""command dispatch :: any -> any
                        refers python:caty.shell.script.node.Dispatch;
    """

    def __init__(self):
        Builtin.__init__(self, None)
        self.__cases = {}
        self.__query = build_query('$')
        self.__scalar_tag_map = {
            types.IntType: ['number'],
            types.FloatType: ['number'],
            types.UnicodeType: ['string'], 
            types.BooleanType: ['boolean'], 
            types.NoneType: ['null'], 
            types.ListType: ['array', 'bag'],
            types.DictType: ['object'],
        }
    
    def add_case(self, case):
        self.__cases[case.tag] = case

    def set_facility(self, facilities):
        for k ,v in self.__cases:
            v.set_facility(facilities)

    def execute(self, opts):
        jsobj = self.input
        target = self.__query.find(jsobj).next()
        if not isinstance(target, TaggedValue) and not (isinstance(target, dict) and '$tag' in target):
            return self.not_tagged_value_case(target)
        else:
            return self.tagged_value_case(target)

    def tagged_value_case(self, target):
        tag = target.tag if isinstance(target, TaggedValue) else target['$tag']
        if tag not in self.__cases:
            raise ScriptError(tag)
        return self.exec_cmd(tag, target)

    def not_tagged_value_case(self, target):
        tags = self.__scalar_tag_map.get(type(target), None)
        if tags == None:
            raise ScriptError()
        for tag in tags:
            if tag in self.__cases:
                return self.exec_cmd(tag, target)
    
    def exec_cmd(self, tag, jsobj):
        childcmd = self.__cases[tag].cmd
        if childcmd.in_schema != jstypes.never:
            childcmd.input = jsobj
        ret = childcmd()
        self.output = childcmd.output



"""
class Switch(Dispatch):
    in_schema_str = u'any'
    out_schema_str = u'any'
    args_schema_str = u'never'
    opts_schema_str = u'never'

    def __init__(self, *args, **kwds):
        Builtin.__init__(self, *args, **kwds)
        self.__cases = []

    def execute(self, opts):
        target = self.__query.find(jsobj).next()
        if target not in self.__cases:
            raise ScriptError(str(target))
        childcmd = self.__cases[target.tag].cmd
        childcmd.input = jsobj
        ret = childcmd()
        self.output = childcmd.output
        return ret
"""
class Case(object):
    def __init__(self, tag, cmd):
        self.tag= tag
        self.cmd = cmd

    def set_facility(self, facilities):
        self.cmd.set_facility(facilities)

