#coding:utf-8
from topdown import *
from caty.shell.script.node import *
from caty.shell.script.builder import combine
from caty.util import bind2nd

class ContinueLine(Exception):
    u"""行継続のシグナル
    """
    pass
class NothingTodo(Exception):
    u"""コメントのみの入力など、何もしないときのシグナル
    """
    def __init__(self):
        pass

class ScriptParser(object):
    def __init__(self, text, builder):
        self.text = text
        self.builder = builder

    def parse(self):
        if not self.text:
            return None
        seq = CharSeq(self.text, hook=self.remove_comment, auto_remove_ws=True)
        parsers = [
                    try_(self.cond),
                    try_(self.object), 
                    try_(self.diagram_order), 
                    try_(self.value), 
                    try_(self.command),
                    try_(self.list), 
                    ]
        try:
            self._script = self.make_pipeline(parsers, seq)
            if not seq.eof:
                raise ParseFailed(seq, parsers)
        except EndOfBuffer, e:
            return None
        return self._script

    def value(self, cs):
        v = cs.parse([self.string, 
                      bind2nd(self.null, True), 
                      bind2nd(self.number, True), 
                      bind2nd(self.integer, True), 
                      bind2nd(self.boolean, True)])
        s = ScalarBuilder()
        s.set_value(v)
        return s

    @staticmethod
    @try_
    def string(seq):
        try:
            _ = seq.parse('"')
            s = seq.parse(until('"'))
            _ = seq.parse('"')
            return s
        except EndOfBuffer, e:
            raise ParseFailed(seq, ScriptParser.string)

    def boolean(self, seq, convert=False):
        b = seq.parse(['true', 'false'])
        if convert:
            return True if b == 'true' else False
        else:
            return b

    def integer(self, seq, convert=False):
        i = seq.parse(Regex(r'-?[0-9]+'))
        if convert:
            return int(i)
        else:
            return i

    def number(self, seq, convert=False):
        n = seq.parse([Regex(r'-?[0-9]+\.[0-9]+([eE][-+]?[0-9]+)?'), Regex(r'-?[0-9]+[eE][-+]?[0-9]+')])
        if convert:
            return float(n)
        else:
            return n

    def null(self, seq, convert=False):
        n = seq.parse('null')
        if convert:
            return None
        else:
            return n

    def name(self, seq):
        return seq.parse(Regex(r'[a-zA-Z]+[-a-zA-Z0-9_]*'))

    def command(self, seq):
        name = seq.parse(self.name)
        opts = seq.parse(many(self.opt))
        o = []
        if opts:
            for op, val in opts:
                o.append(op)
                if val:
                    o.append(val)
        args = seq.parse(many(self.arg))
        if not args:
            args = []
        return self.builder.build([name] + o + args)

    def pipe(self, seq):
        _ = seq.parse('|')
        return _

    def make_pipeline(self, parsers, seq):
        def _pipeline(s):
            r = []
            s.parse(self.do_nothing)
            if s.eof:
                raise NothingTodo()
            r.append(s.parse(parsers))
            while True:
                a = option(self.pipe, None)(s)
                if not a:
                    break
                if s.eof:
                    raise ContinueLine()
                r.append(s.parse(parsers))
            return combine(r)

        def _group(s):
            _ = s.parse('(')
            r = _pipeline(s)
            _ = s.parse(')')
            return r

        def _tag(s):
            _ = s.parse('@')
            n = s.parse(self.name)
            r = s.parse([_pipeline, _group])
            return TagBuilder(n, r)

        return seq.parse([_pipeline, _group])

    def unquoted(self, seq):
        return seq.parse(Regex(r'[^-\t\r\n <>|+"(){},\[\]]+'))

    def opt(self, seq):
        o = seq.parse([self.shortopt, self.longopt])
        v = seq.parse([self.integer, self.number, self.boolean, self.unquoted, self.string])
        return o, v

    def shortopt(self, seq):
        return seq.parse(Regex(r'-[a-zA-Z]'))

    def longopt(self, seq):
        return seq.parse(Regex(r'--[a-zA-Z]+[a-zA-Z0-9]*'))

    def arg(self, seq):
        return seq.parse([self.integer, self.number, self.boolean, self.unquoted, self.string])
    
    def comma(self, seq):
        _ = seq.parse(',')
        return _

    def list(self, seq):
        seq.parse('[')
        values = seq.parse(split(self.listitem, self.comma, True))
        seq.parse(']')
        l = ListBuilder()
        l.set_values(anything(values))
        return l

    def listitem(self, seq):
        if seq.current == ']': return
        parsers = [
                    try_(self.cond),
                    try_(self.object), 
                    try_(self.diagram_order), 
                    try_(self.value), 
                    try_(self.command),
                    try_(self.list), 
                    ]

        try:
            return self.make_pipeline(parsers, seq)
        except NothingTodo:
            raise ParseFailed(seq, self.listitem)

    def object(self, seq):
        seq.parse('{')
        items = seq.parse(split(self.item, self.comma, True))
        seq.parse('}')
        o = ObjectBuilder()
        for i in anything(items):
            o.add_node(i)
        return o

    def item(self, seq):
        if seq.current == '}': return
        n = self.string(seq)
        seq.parse(':')
        try:
            v = self.make_pipeline([self.value, 
                                    self.command, 
                                    self.list, 
                                    self.object, 
                                    self.diagram_order], seq)
        except NothingTodo:
            raise ParseFailed(seq, self.item)
        return CommandNode(n, v)

    def diagram_order(self, seq):
        seq.parse('do')
        seq.parse('{')
        items = seq.parse(split(self.item_do, self.comma))
        seq.parse('}')
        o = ObjectBuilder()
        for i in anything(items):
            o.add_node(i)
        return o

    def item_do(self, seq):
        if seq.current == '}': return
        try:
            v = self.make_pipeline([self.value, 
                                    self.command, 
                                    self.list, 
                                    self.object, 
                                    self.diagram_order], seq)
        except NothingTodo:
            raise ParseFailed(seq, self.item)
        seq.parse('>:')
        n = self.string(seq)
        return CommandNode(n, v)

    def cond(self, seq):
        seq.parse('when')
        seq.parse('{')
        cases = seq.parse(split(self.case, self.comma))
        seq.parse('}')
        d = Dispatch()
        for c in anything(cases):
            d.add_case(c)
        return d

    def case(self, seq):
        if seq.current == '}': return
        t = self.string(seq)
        seq.parse('=>')
        try:
            v = self.make_pipeline([self.value, self.command, self.list, self.object, self.diagram_order], seq)
        except NothingTodo:
            raise ParseFailed(seq, self.case)
        return Case(t, v)

    def remove_comment(self, seq):
        seq.parse([self.singleline_comment, self.multiline_comment])

    def singleline_comment(self, seq):
        seq.parse('//')
        seq.parse([until('\n'), until('\r')])

    def multiline_comment(self, seq):
        seq.parse('/*')
        seq.parse(until('*/'))
        seq.parse('*/')

    def do_nothing(self, seq):
        u"""コメント行のみ入力された場合用のダミー
        """
        pass


def anything(l):
    return filter(lambda i: i!= None, l)
