# coding: utf-8
from __future__ import with_statement
from caty.jsontools import json
from caty.jsontools.path import build_query
from threading import RLock
from hashlib import sha1
import random
import sys

__all__ = ['SessionValue', 'SessionStorage', 'Session', 'get_session']
FORBIDDEN = set(['grant', 'loggedin'])
PRIVILEDE = {
    'login': set(['grant', 'loggedin'])
}
class SessionStorage(object):
    u"""セッションデータ一覧を保持し、セッション情報の取得と永続化を行う。
    """
    __sessions = {}
    
    @classmethod
    def create(self):
        from caty.shell.system import SECRET_KEY
        u"""新しいセッション情報を生成する。
        """
        while 1:
            with RLock():
                key = sha1(reduce(lambda a, b: a + b,
                                        [str(random.randint(0, sys.maxint - 1)),
                                        str(random.randint(0, sys.maxint - 1)),
                                        SECRET_KEY])).hexdigest()
                if key not in self.__sessions:
                    break
        self.__sessions[key] = Session(key)
        return self.__sessions[key]

    def save(self, session):
        pass

    @classmethod
    def restore(self, session_key):
        return self.__sessions.get(session_key, None)

    def delete(self, session_key):
        pass

class SessionValue(object):
    def __init__(self, json_type, value, addable=False):
        self.__json_type = json_type
        self.__value = value
        self.__addable = addable

    @property
    def json_type(self):
        return self.__json_type

    @property
    def value(self):
        return self.__value

    @property
    def addable(self):
        return self.__addable

    def _to_dict(self):
        u"""JSON データとして永続化する際に用いる。
        フレームワーク側のみが使用し、基本的にアプリケーションからは使用しない。
        """
        return {
            'json_type': self.json_type,
            'value': self.value,
            'addable': self.addable,
        }

class Session(object):
    u"""セッション情報。
    任意の情報をセッションに格納できる。
    ただし、一度格納した値は上書き不能となる。
    put時にオプションを指定することで、リストに対する値の追加のみ可能となる。
    """

    def __init__(self, key):
        self.__key = key
        self.__session_obj = {}
        self.__handles = {}

    @property
    def key(self):
        return self.__key

    def exists(self, name):
        u"""name に対応した値が格納されている場合 True を返す。
        """
        return name in self.__session_obj
    
    def addable(self, name):
        u"""name が追記可能オブジェクトであれば True を返す。
        """
        if self.exists(name):
            return self.__session_obj[name].addable
        raise KeyError(name)

    def get(self, name):
        u"""name に対応した値を返す。
        """
        return self.__session_obj[name].value

    def find(self, name, sjpath):
        u"""SJPath を用いて値を返す。
        """
        return build_query(sjpath).find(self.get(name))

    def put(self, name, json_type, value, addable=False):
        u"""値の格納。値が存在する、あるいは予約されている場合は例外を送出する。
        """
        if self.exists(name) or self.reserved(name): raise Exception('Can not overwrite existing value: %s' % name)
        if addable and not isinstance(value, list): raise Exception('addable is not able to apply to "%s" instance.' % str(value.__class__))
        if name in FORBIDDEN: raise Exception('Reserved for privileged command: %s' % name)
        json_type.validate(value)
        self.__session_obj[name] = SessionValue(json_type, value, addable)
        return

    def put_privilege(self, registrar, name, json_type, value):
        cmd = registrar.__class__.__name__.lower()
        if cmd not in PRIVILEDE:
            raise Exception('Not privileged command: command=%s, key=%s' % (cmd, name))
        type_list = PRIVILEDE[cmd]
        if name not in type_list:
            raise Exception('Not privileged command: %s' % (cmd, name))
        json_type.validate(value)
        self.__session_obj[name] = SessionValue(json_type, value)
        return

    def add(self, name, value):
        u"""値の追記。 put, reserve 時に addable オプションに True を指定された場合のみ可能。
        """
        if not self.addable(name):
            raise Exception('Operation not allowed for "%s"' % name)
        self.get(name).append(value)
        return

    def reserve(self, name, json_type, addable=False):
        u"""あらかじめ name を予約し、別の場所で値を追加されないようにする。
        """
        if name in self.__handles:
            raise Exception('Already reserved: %s' % name)
        if self.exists(name):
            raise Exception('Already exists: %s' % name)
        handle = 'handle_%s' % name
        self.__handles[name] = (name, json_type, addable)
        return

    def reserved(self, name):
        u"""name が予約済みであれば True を返す。
        """
        return name in self.__handles

    def set(self, handle, value):
        u"""handle によって予約済みの領域に値を格納する。
        """
        if not handle in self.__handles:
            raise Exception('Not reserved handle')
        name, json_type, addable = self.__handles[handle]
        json_type.value(value)
        self.put(name, json_type, value, addable)
        return

    def clear(self, clear_reserved_handles=False):
        u"""セッション変数をすべて削除する。
        clear_reserved_handles が True の場合、予約している領域も削除する。
        """
        self.__session_obj.clear()
        if clear_reserved_handles:
            self.__handles.clear()

    def _to_dict(self):
        objects = {}
        for k, v in self.__session_obj.iteritems():
            objects[k] = v._to_dict()
        return {
            'key': self.__key,
            'objects': objects,
        }

