# coding: utf-8
import os
import sys

from caty import mafs, web
from caty.shell import script
from caty.jsontools import schema
from caty.util import memoize
import caty.jsontools as json

DEBUG = True
ROOT_DIR = None
CURRENT_ROOT = None # 対話シェルでのみ使われ、組み込みコマンドの cd により値が更新され得る。
SECRET_KEY = ''
ENCODING = 'utf-8'

def bind1st(fst, f):
    u"""mafs モジュールのクラス/関数は認可トークンを第一引数に必ず持つ。
    認可トークンの存在を shell や template の各関数が意識するのは煩雑に過ぎるので、
    モジュールの構築が終わり実処理が始まった時点で、認可トークンを設定してしまう。
    """
    def binded(*args):
        if args:
            return f(fst, *args)
        else:
            return f(fst)
    return binded

def _delete_common_prefix(src, target):
    common = os.path.commonprefix((src, target))
    return target.replace(common, '')

def comp_slash_file(f, root):
    def _comp(token, path, mode='rb'):
        if path.startswith('/'):
            return f(token, path, mode)
        else:
            p = _delete_common_prefix(root, os.path.abspath(path))
            return f(token, '/' + p, mode)
    return _comp

def comp_slash_dir(f, root):
    def _comp(token, path):
        if path.startswith('/'):
            return f(token, path)
        else:
            p = _delete_common_prefix(root, os.path.abspath(path))
            return f(token, '/' + p)
    return _comp

class FSWrapper(object):
    def __init__(self, open_f, file_class, dir_class, glob_f):
        self.__open = open_f
        self.__fo = file_class
        self.__do = dir_class
        self.__glob = glob_f

    def bind_token(self, token):
        o = bind1st(token, self.__open)
        FileObject = bind1st(token, self.__fo)
        DirectoryObject = bind1st(token, self.__do)
        return FSWrapper(o, FileObject, DirectoryObject, None)

    @property
    def open(self):
        return self.__open

    @property
    def FileObject(self):
        return self.__fo

    @property
    def DirectoryObject(self):
        return self.__do


import __builtin__
import locale
@memoize
def setup_caty(config_file_path, interactive=False):
    u"""Caty 環境のセットアップを行う。
    主に mafs の初期化処理が該当する。
    この関数は同一の設定ファイルであればプロセス一つにつき一回呼ばれれば良く、
    設定ファイルの更新への追従などは現時点で考えていない。
    呼び出しが一回限りで済むことはメモ化によって保証する。
    戻り値は mafs へのラッパーと Web サーバの設定オブジェクト（設定ファイルに記述があれば）となる。
    
    interactive が True の場合、パス名に / が自動的に補完される。
    これは対話シェルでの使い勝手の向上のためである。
    対話シェル以外の環境では、混乱を防ぐためにデフォルトの False のままとすること。
    """
    cfg = json.load(__builtin__.open(config_file_path))
    fs = cfg['mafs']['filesystem']
    root_dir = cfg['mafs']['root_directory']
    exec 'import %s as mafs_impl' % fs
    fo = mafs.initialize(fs=mafs_impl, root_dir=root_dir)
    global DEBUG
    global SECRET_KEY
    global ENCODING
    open = fo.open
    FileObject = fo.FileObject
    DirectoryObject = fo.DirectoryObject
    if 'debug' in cfg:
        DEBUG = cfg['debug']
    if 'secret_key' in cfg:
        SECRET_KEY = cfg['secret_key']
    if 'mime_types' in cfg['mafs']:
        from caty.mafs import metadata
        metadata.MIME_TYPES.update(cfg['mafs']['mime_types'])
    ROOT_DIR = root_dir
    if interactive:
        open = comp_slash_file(open, root_dir)
        FileObject = comp_slash_file(FileObject, root_dir)
        DirectoryObject = comp_slash_dir(DirectoryObject, root_dir)
        ENCODING = locale.getpreferredencoding()
    if 'web' in cfg:
        web.initialize(cfg['web'])
    fs = FSWrapper(open, FileObject, DirectoryObject, None) # TODO: glob 関数を作ること
    mafs.filesystem = fs
    # スキーマモジュールの初期化時には、常に管理者権限で動作する。
    # そうでないと権限の関係上、スキーマ定義ファイルが読み込めない
    schema.initialize(fs.bind_token(mafs.authorization.AuthoriToken('admin')))
    # スクリプトモジュールの初期化
    try:
        sys.path.append(root_dir + '/commands')
    except ImportError, e:
        print 'Warning: There is no "commands" module in caty file system'
    script.initialize(schema.registrar)

import codecs
import locale
class StdStream(object):
    def __init__(self): #, encoding):
        self.encoding = locale.getpreferredencoding()
        self.out = codecs.getwriter(locale.getpreferredencoding())(sys.stdout)
        self.input = codecs.getwriter(locale.getpreferredencoding())(sys.stdin)

    def write(self, content):
        self.out.write(content)

    def read(self):
        return json.loads(self.input.read())

class Environment(object):
    u"""環境クラス。
    環境変数、入出力ストリームなどを扱う。
    環境変数の実体は caty.shell.session で定義される Session クラスが想定されるが、
    同様なインターフェースであれば特に問題は無い。
    Session クラスはそれぞれセッションスコープとリクエストスコープに分けられ、
    前者はアプリケーションが終了するまで（あるいは永続化を行いそれ以降も）生存するデータであり、
    後者は一連のコマンドのパイプラインの間という短期のセッションに対応するデータである。
    """
    def __init__(self, session_scope_obj, request_scope_obj):
        self.__session = session_scope_obj
        self.__request = request_scope_obj
        self.stream = StdStream()
        self.session_timeout = ''
        self.server_url = ''
        self.login_path = ''
    
    @property
    def root_dir(self):
        return ROOT_DIR

    def session():
        def _get(self):
            return self.__session

        def _set(self, v):
            self.__session = v
        return _get, _set
    session = property(*session())

    def request():
        def _get(self):
            return self.__request

        def _set(self, v):
            self.__request = v
        return _get, _set
    request = property(*request())

