# coding: utf-8
import sys
import os
sys.path.insert(0, os.path.abspath('../../../../'))
import unittest
from caty.template.core.vm import *
from caty.template.core.loader import IBytecodeLoader

class MockLoader(IBytecodeLoader):
    u"""テスト用バイトコードローダー。
    あらかじめパス名とバイトコードを与えておき、
    load メソッドでは常にそのバイトコードを使い、
    store メソッドでは何もしない。
    """
    def __init__(self, *args, **kwds):
        self.codes = {}

    def store(self, path):
        pass

    def set_bytecode(self, path, code):
        u"""code は (opcode, value) のタプルのリストでもよい。
        """
        self.codes[path] = code

    def load(self, path):
        return self.codes[path]

class ContextTest(unittest.TestCase):
    def testDefaultContext(self):
        u"""通常のアクセスのテスト
        """
        c = Context({
            'a': 'foo',
            'b': 'bar'
        })
        self.assertEquals('foo', c['a'])
        self.assertEquals('bar', c['b'])

    def testHierarcalAccess(self):
        u""".による階層アクセスのテスト
        """
        c = Context({
            'a': {
                'b': 'boo',
                'c': {
                    'd': 'damn',
                }
            }
        }
        )
        self.assertEquals('boo', c['a.b'])
        self.assertEquals('damn', c['a.c.d'])

    def testShadowing(self):
        u"""複数の辞書で同じキーが定義されてる場合のテスト
        最も直近に追加されたコンテキストが参照されるのが正しい
        """
        c = Context({'a': 'foo'})
        c.new({})
        c['a'] = 'bar'
        self.assertEquals('bar', c['a'])

    def testNew(self):
        u"""新規の辞書をコンテキストに追加する
        """
        c = Context({'a': 'foo'})
        c.new({'b': 'bar'})
        self.assertEquals('bar', c['b'])
        self.assertEquals(2, len(c.dict))

    def testDelete(self):
        u"""コンテキストから辞書を削除
        """
        c = Context({'a': 'foo'})
        c.new({'b': 'bar'})
        c.delete()
        self.assertEquals(1, len(c.dict))

    def testCannotAddingToDefaultContext(self):
        u"""デフォルトの辞書への追加操作が失敗するかのテスト
        """
        c = Context({'a': 'foo'})
        try:
            c['b'] = 'bar'
            self.fail()
        except TemplateRuntimeError, e:
            pass
        except:
            self.fail()

    def testCannotDeleteDefaultContext(self):
        u"""デフォルトの辞書が削除されないことのテスト
        """
        c = Context({'a': 'foo'})
        try:
            c.delete()
            self.fail()
        except TemplateRuntimeError, e:
            pass
        except:
            self.fail()

    def testAddingHierarcalData(self):
        u""".で連結された階層キーのセット
        """
        c = Context({'a': 'foo'})
        c.new({})
        c['x.y.z'] = 'buz'
        c['x.y.a'] = 'buz?'
        self.assertEquals('buz', c['x.y.z'])
        self.assertEquals('buz?', c['x.y.a'])
        self.assertTrue(isinstance(c.dict[0]['x'], dict))
        self.assertTrue(isinstance(c.dict[0]['x']['y'], dict))
        self.assertFalse(isinstance(c.dict[0]['x']['y']['z'], dict))

    def testGet(self):
        u"""辞書の同メソッドと同じ動作かのテスト
        """
        c = Context({'a': 'foo'})
        v = c.get('x', 0)
        self.assertEquals(0, v)

class VirtualMachineTest(unittest.TestCase):
    def testVariableReferrence(self):
        u"""変数参照のテスト
        """
        i = VirtualMachine(MockLoader())
        c = {'a': 'foo'}
        b = (
            (LOAD, 'a'),
            (POP, None),
        )
        i.context = Context(c)
        self.assertEquals('foo', i.evaluate(b))

    def testPush(self):
        u"""無名の値の表示テスト
        """
        i = VirtualMachine(MockLoader())
        b = (
            (PUSH, 'a'),
            (POP, None),
            (PUSH, 10),
            (POP, None),
        )
        self.assertEquals('a10', i.evaluate(b))

    def testJmp(self):
        u"""単純なラベル定義とジャンプのテスト
        """
        i = VirtualMachine(MockLoader())
        b = (
            (PUSH, 'a'),
            (POP, None),
            (JMP, 'label'),
            (PUSH, 10),
            (POP, None),
            (LABEL, 'label'),
            (PUSH, 'b'),
            (POP, None),
        )
        self.assertEquals('ab', i.evaluate(b))

    def testJmpUnless(self):
        u"""条件付きジャンプのテスト
        """
        i = VirtualMachine(MockLoader())
        b = (
            (PUSH, 'a'),
            (POP, None),
            (PUSH, True),
            (JMPUNLESS, 'label'),
            (PUSH, 10),
            (POP, None),
            (JMP, 'true'),
            (LABEL, 'label'),
            (PUSH, 'b'),
            (POP, None),
            (LABEL, 'true'),
            (PUSH, 'c'),
            (POP, None),
        )
        self.assertEquals('a10c', i.evaluate(b))

    def testNewCtx(self):
        u"""コンテキストリストの伸長のテスト
        """
        i = VirtualMachine(MockLoader())
        c = {'a': 'foo'}
        i.context = c
        b = (
            (NEWCTX, None),
        )
        i.evaluate(b)
        self.assertEquals(len(i.context.dict), 2)

    def testDelCtx(self):
        u"""コンテキストリストの削除のテスト
        """
        i = VirtualMachine(MockLoader())
        c = {'a': 'foo'}
        i.context = c
        b = (
            (NEWCTX, None),
            (DELCTX, None),
            (LOAD, 'a'),
            (POP, None)
        )
        self.assertEquals('foo', i.evaluate(b))
        self.assertEquals(len(i.context.dict), 1)
        
    def testCPush(self):
        u"""コンテキストに値を追加する
        """
        i = VirtualMachine(MockLoader())
        c = {'a': 'foo'}
        i.context = c
        b = (
            (NEWCTX, None),
            (PUSH, 'bar'),
            (CPUSH, 'a'),
            (LOAD, 'a'),
            (POP, None)
        )
        self.assertEquals('bar', i.evaluate(b))

    def testMacro(self):
        u"""マクロを呼び出す
        """
        i = VirtualMachine(MockLoader())
        c = {'a': 'funga'}
        i.context = c
        b = (
            (MACRO, 'macro1'),
            (LOAD, 'arg1'),
            (POP, None),
            (LOAD, 'arg2'),
            (POP, None),
            (RETURN, None),
            (NEWCTX, None),
            (PUSH, 'hoge'),
            (CPUSH, 'arg1'),
            (PUSH, 'fuga'),
            (CPUSH, 'arg2'),
            (EXPAND, 'macro1'),
            (LOAD, 'a'),
            (POP, None),
        )
        self.assertEquals('hogefugafunga', i.evaluate(b))

    def testLoop(self):
        u"""ループ処理のテスト
        ジャンプと比較演算などでループをエミュレート
        """
        i = VirtualMachine(MockLoader())
        c = {'l': ['foo', 'bar', 'buz']}
        i.context = c
        b = (
            (STRING, 'loop start'),
            (NEWCTX, None),
            (PUSH, 0),
            (CPUSH, 'index'),
            (PUSH, 1),
            (CPUSH, 'iteration'),
            (LEN, 'l'),
            (CPUSH, 'count'),
            
            (LABEL, 'loop'), # ループ開始
            
            (PUSH, 0),
            (LOAD, 'index'),
            (EQ, None),
            (CPUSH, 'first'),

            (LOAD, 'count'),
            (LOAD, 'iteration'),
            (EQ, None),
            (CPUSH, 'last'),
            
            (STRING, '\ndata = '),
            (LOAD, 'l'),
            (LOAD, 'index'),
            (AT, None),
            (POP, None),
            (STRING, '\n  index = '),
            (LOAD, 'index'),
            (POP, None),
            (STRING, '\n  iteration = '),
            (LOAD, 'iteration'),
            (POP, None),
            (STRING, '\n  first ? '),
            (LOAD, 'first'),
            (POP, None),
            (STRING, '\n  last ? '),
            (LOAD, 'last'),
            (POP, None),

            # ループ変数の更新
            (LOAD, 'index'),
            (PUSH, 1),
            (ADD, None),
            (CPUSH, 'index'),
            (LOAD, 'iteration'),
            (PUSH, 1),
            (ADD, None),
            (CPUSH, 'iteration'),

            (LOAD, 'last'),
            (JMPUNLESS, 'loop'),
            (DELCTX, None),
            (STRING, '\nloop end\n')
        )
        expected = """loop start
data = foo
  index = 0
  iteration = 1
  first ? True
  last ? False
data = bar
  index = 1
  iteration = 2
  first ? False
  last ? False
data = buz
  index = 2
  iteration = 3
  first ? False
  last ? True
loop end
"""
        result = i.evaluate(b)
        self.assertEquals(expected, result)



if __name__ == '__main__': 
    result = unittest.TestResult()
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(ContextTest))
    suite.addTest(unittest.makeSuite(VirtualMachineTest))
    suite.run(result)
    if not result.wasSuccessful():
        print '--------'.join(('%s:\n%s' % (f[0]._testMethodName, f[1]) for f in result.failures + result.errors))

