# coding: utf-8
import types
from StringIO import StringIO

from caty.template.core.instructions import *
from caty.template.core.filter import *
from caty.util import merge_dict

class StringWrapper(unicode):
    def __init__(self, s):
        if isinstance(s, StringWrapper):
            self.string = s.string
        else:
            self.string = s

class RawString(StringWrapper):
    u"""Unicode データのラッパークラス。
    出力ストリームに書き出す際にエスケープしない。
    """

    def __str__(self):
        return self.string

class HTMLString(StringWrapper):
    u"""Unicode データのラッパークラス。
    出力ストリームに書き出す際に HTML 特殊文字のエスケープを行う。
    """

    def encode(self, encoding):
        return self.__str__(str.encode(self, encoding))

    def __str__(self):
        return self.string.replace('&', '&amp;').replace('"', '&quot;').replace('<', '&lt;').replace('>', '&gt;').replace('\'', '&#39;')

class RawStringFilter(AbstractFilter):
    name = 'raw'
    def filter(self, s):
        if isinstance(s, HTMLString):
            return RawString(s.string)
        else:
            return RawString(s if isinstance(s, basestring) else str(s))

class TemplateRuntimeError(Exception): pass

class BytecodeIterator(object):
    u"""バイトコード用のコンテナクラス。
    イテレータのように使用可能だが、ジャンプ命令に対応するためにランダムアクセス可能とする。
    あくまでもイテレータなので、要素への変更は行わない事を前提とする。
    最適化やマクロによるバイトコード変換時には、新たにイテレータを再生成すること。
    """
    def __init__(self, iterable):
        u"""任意のイテレート可能オブジェクトを引数とする。
        """
        self.iterable = tuple(iterable)
        self.cursor = 0
        self.last = len(self.iterable)

    def next(self):
        if self.last - 1 <= self.cursor:
            raise StopIteration
        r = self.iterable[self.cursor]
        self.cursor += 1
        return r

    index = property(lambda self:self.cursor)

    def jump_to(self, index):
        u"""指定されたインデックスにイテレータの現在位置を設定する。
        """
        self.cursor = index
        if self.cursor < -1:
            self.cursor = -1

    def _current(self):
        return self.iterable[self.cursor]
    current = property(_current)

class Undefined(object):
    def __nonzero__(self):
        return False

UNDEF = Undefined()

class Context(object):
    u"""変数参照用の名前空間。
    辞書に対するラッパーであり、 a.b.c のような階層アクセスを実装する。
    また、ラップ対象の辞書は単一の辞書ではなく辞書のリストであり、
    複数の辞書に同一のキーが存在する場合、最初に見つかったものを返すこととする。
    これは for-each に類する構文の実現のために用いる。
    """
    def __init__(self, dictionaly):
        self.dict = [dictionaly]

    def __getitem__(self, key):
        keys = key.split('.')
        return self._rec_find(keys, self.dict)

    def _rec_find(self, keys, dictionaries):
        k = keys.pop(0)
        for d in dictionaries:
            if k in d:
                if len(keys) == 0:
                    return d[k]
                else:
                    n = d[k]
                    if not isinstance(n, list):
                        return self._rec_find(keys, [n])
                    else:
                        return self._rec_find(keys, n)
        raise TemplateRuntimeError('Undefined variable: %s' % '.'.join(keys))

    def get(self, *args):
        key = args[0]
        try:
            return self[key]
        except:
            if len(args) > 1:
                return args[1]
            else:
                raise
                
    def new(self, new_context):
        u"""コンテキストに新たな辞書を追加する。
        もしもコンテキストが空だった場合、何もしない。
        """
        self.dict.insert(0, new_context)

    def is_context_list_length_gt1(f):
        u"""コンテキストの内包する辞書のリストが1以上でなかった場合エラーとする。
        """
        def _(self, *args, **kwds):
            if len(self.dict) < 2:
                raise TemplateRuntimeError('Can not modify default context')
            return f(self, *args, **kwds)
        return _

    @is_context_list_length_gt1
    def delete(self):
        u"""コンテキストの辞書リストの先頭を削除する。
        """
        self.dict.pop(0)

    @is_context_list_length_gt1
    def __setitem__(self, name, value):
        u"""辞書リストの先頭の辞書に新たな要素を追加する。
        a.b.c のような階層化された name が与えられた場合、
        それらは a: {b: c: value}} という辞書に変換される。
        """
        d = reduce(lambda x, y: {y: x}, reversed(name.split('.') + [value]))
        if d.keys()[0] not in self.dict[0] or not isinstance(d.values()[0], dict):
            self.dict[0][d.keys()[0]] = d.values()[0]
        else:
            merge_dict(self.dict[0][d.keys()[0]], d.values()[0])
    
    def merge(self, newdict):
        u"""辞書リストの先頭に新たな辞書をマージする。
        """
        merge_dict(self.dict[0], newdict)
    
    def __str__(self):
        return '\n'.join(str(i) for i in self.dict)

class Bytecode(object):
    u"""バイトコードオブジェクト。
    別にタプルでも良いのだが、可読性のためにこちらを採用。
    """
    def __init__(self, opcode, value=''):
        self.opcode = opcode
        self.value = value

    def __str__(self):
        return ('%s %s' % (get_name(self.opcode), str(self.value)))

class Stack(object):
    u"""リストベースのスタック。
    リストをそのまま使ってもよいが、 append/pop よりも
    push/pop/top/ のようなインターフェースの方が
    スタックマシンを操作するときのメンタルモデルに近い。
    """
    def __init__(self):
        self.list = []

    def push(self, v):
        self.list.append(v)

    def pop(self):
        try:
            return self.list.pop(-1)
        except IndexError:
            raise TemplateRuntimeError('Stack is empty')

    def top(self):
        return self.list[-1]

    def is_empty(self):
        return len(self.list) == 0

class VirtualMachine(object):
    u"""バイトコードを解釈し、出力ストリームに書き出す仮想マシン。
    あくまでも処理の対象はバイトコードであり、
    ファイル/ストリームからバイトコードを読み出したりコンパイルしたりという作業は別クラスへ委譲する。
    """
    def __init__(self, bytecode_loader):
        self._context = Context({})
        self._filters = FilterContainer()
        self._filters.add(RawStringFilter())
        self._filters.add(NotNullFilter())
        self._filters.add(NotEqFilter())
        self._filters.add(EqFilter())
        self._filters.add(UpperFilter())
        self._filters.add(LowerFilter())
        self._macro = {}
        self._out = StringIO()
        self._bytecode_loader = bytecode_loader
        self._include_callback = None

    def filters():
        def get(self):
            return self._filter

        def set(self, value):
            self._filter = value
        return get, set
    filters = property(*filters())

    def context():
        def get(self):
            return self._context

        def set(self, value):
            if isinstance(value, dict):
                self._context = Context(value)
            else:
                self._context = value

        return get, set
    context = property(*context())

    def dup(self):
        new_vm = VirtualMachine(self._bytecode_loader)
        new_vm._context = self._context
        new_vm._filters = self._filters
        new_vm._out = self._out
        return new_vm

    def _evaluate(self, bytecode):
        u"""バイトコードの実行関数。
        バイトコードのリストを受け取り、
        実行した結果をジェネレータとして返す。
        """
        if not isinstance(bytecode, BytecodeIterator) and isinstance(bytecode[0], Bytecode):
            bi = BytecodeIterator(bytecode)
        elif not isinstance(bytecode[0], Bytecode):
            bi = BytecodeIterator([Bytecode(*b) for b in bytecode])
        else:
            bi = bytecode
        stack = Stack()
        context = self.context
        labelmap = {}
        opmap = {}
        filters = self._filters
        def load(k):
            stack.push(context.get(k))
            return
        opmap[LOAD] = load
        
        def push(v):
            stack.push(v)
            return
        opmap[PUSH] = push

        def pop(ignore):
            s = stack.pop()
            if isinstance(s, RawString):
                return s
            elif isinstance(s, basestring):
                return HTMLString(s)
            else:
                return s
        opmap[POP] = pop

        def label(name):
            labelmap[name] = bi.index
            return
        opmap[LABEL] = label

        def jmp(l):
            index = labelmap.get(l, -1)
            if index == -1:
                while bi.current:
                    code = bi.current.opcode
                    value = bi.current.value
                    if code == LABEL and value == l:
                        label(l)
                        break
                    bi.next()
                else:
                    raise TemplateRuntimeError('Undefined label: %s' % l)
            else:
                bi.jump_to(index)
            return
        opmap[JMP] = jmp

        def jmpunless(l):
            if not stack.pop():
                jmp(l)
            return
        opmap[JMPUNLESS] = jmpunless

        def newctx(ignore):
            context.new({})
            return
        opmap[NEWCTX] = newctx

        def delctx(ignore):
            context.delete()
        opmap[DELCTX] = delctx

        def cpush(k):
            context[k] = stack.pop()
            return
        opmap[CPUSH] = cpush

        def macro(name):
            if name in self._macro:
                return
            i = bi.index
            while bi.current:
                if bi.current.opcode == RETURN:
                    break
                bi.next()
            else:
                TemplateRuntimeError('Missing end of macro definition: %s' % name)
            self._macro[name] = i
        opmap[MACRO] = macro
        
        def ret(ignore):
            bi.jump_to(stack.pop())
            return
        opmap[RETURN] = ret

        def expand(name):
            macro_index = self._macro.get(name, None)
            if not macro_index:
                TemplateRuntimeError('Macro definition not found: %s' % name)
            push(bi.index)
            bi.jump_to(macro_index)
            return
        opmap[EXPAND] = expand

        def string(value):
            return RawString(value)
        opmap[STRING] = string

        def defined(key):
            if UNDEF == context.get(key, UNDEF):
                push(False)
            else:
                push(True)
        opmap[DEFINED] = defined

        def call(args):
            if isinstance(args, list):
                name, argc = args
            else:
                name = args
                argc = 1
            argc = int(argc)
            argv = []
            for i in range(argc):
                argv.append(pop(None))
            argv = list(reversed(argv))
            f = filters.get(name)
            if f == None:
                raise TemplateRuntimeError('Undefined filter: %s' % name)
            push(f(*argv))
            return
        opmap[CALL] = call
        
        def include(name):
            if self._include_callback == None:
                newvm = self.dup()
                newcode = self._bytecode_loader.load(name)
                newvm.context.new({})
                newvm.write(newcode, self._out)
                newvm.context.delete()
            else:
                self._out.write(self._include_callback(name))
            return
        opmap[INCLUDE] = include

        def negate(ignore):
            stack.push(not stack.pop())
            return
        opmap[NOT] = negate

        def add(ignore):
            stack.push(stack.pop() + stack.pop())
            return
        opmap[ADD] = add

        def eq(ignore):
            stack.push(stack.pop() == stack.pop())
            return
        opmap[EQ] = eq

        def lt(ignore):
            stack.push(stack.pop() < stack.pop())
            return
        opmap[LT] = eq

        def le(ignore):
            stack.push(stack.pop <= stack.pop())
            return
        opmap[LE] = le
        
        def length(key):
            stack.push(len(context[key]))
            return
        opmap[LEN] = length

        def at(ignore):
            i = stack.pop()
            ls = stack.pop()
            stack.push(ls[i])
            return 
        opmap[AT] = at

        def ref(ignore):
            stack.push(context.dict)
            return
        opmap[REFCONTEXT] = ref

        while bi.current:
            code = bi.current.opcode
            value = bi.current.value
            operation = opmap.get(code, None)
            if not operation:
                raise TemplateRuntimeError('Undefined operation: %s' % code)
            r = operation(value)
            if r != None:
                if isinstance(r, types.GeneratorType):
                    for ri in r:
                        yield r
                else:
                    yield r
            bi.next()
    
    def evaluate(self, bytecode):
        return ''.join(str(r) for r in (self._evaluate(bytecode)))

    def write(self, bytecode, out):
        self._out = out
        for r in self._evaluate(bytecode):
            out.write(r if isinstance(r, basestring) else str(r))

