# coding: utf-8
u"""テストスイート起動ユーティリティ。
Makefile にゴチャゴチャ書きすぎるとポータビリティの点で極めて良くなく、
テスト実行用のスクリプトが必要になったので作った。
"""
import tokenize
import os
import subprocess
import sys

def lastmod_of_testcase(target):
    u"""target の内容を解析して、 import してるファイルを割り出し、
    それらのうちもっとも更新時刻の新しいものを target の更新時刻として返す。
    この関数が呼ばれた時点で target の存在するディレクトリに移動しているものとし、
    依存対象のファイルは target と同じディレクトリに存在するものとする。
    """
    def generate_depend_files():
        is_import_line = False
        imported = False
        for tp, s, rc, erc, line in tokenize.generate_tokens(open(target).readline):
            if tp == tokenize.NAME and s == 'import' and not is_import_line:
                is_import_line = True
            elif tp == tokenize.NAME and s == 'from' and not is_import_line:
                is_import_line = True
            elif tp == tokenize.NAME and is_import_line and not imported:
                imported = True
                yield s
            if tp == tokenize.NEWLINE:
                is_import_line = False
                imported = False

    def lastmod_list():
       for f in generate_depend_files():
           if os.path.exists(f + '.py'):
               yield os.stat(f + '.py').st_mtime
       yield os.stat(target).st_mtime
    
    return max(list(lastmod_list()))

def main(target, testsuite=None):
    u"""テストの実行。以下のいずれかの状態の場合にテストを実行する。
    
    * .pyc ファイルよりもソースファイルが新しい
    * ソースファイルと .pyc ファイルのどちらよりもテストの方が新しい

    testsuiteが渡されなかった場合、test/`target` がテストとして使われる。
    テストに失敗した場合、 .pyc ファイルは削除される。
    また、テストファイルは Python スクリプトとして実行可能でなければならず、
    テストが失敗した場合は 0 でない終了ステータスでプロセスを終了させなければならない。
    
    テストケースがカレントディレクトリに存在しない場合、テストケースのあるディレクトリに移動し、
    その後カレントディレクトリに戻るという処理を行う。
    そのため、テストは常にテストケースのあるディレクトリで実行される事を前提にする必要がある。

    テストケースが複数のファイルから成り立つ場合、それらの依存関係を解析し、
    この場合、テストケースを構成する各ファイルはすべてテストケースと同じディレクトリにある必要がある。
    """
    if not testsuite:
        testsuite = 'test/' + target
    pyc = target + 'c'
    curdir = os.getcwd()
    dirname, testname = os.path.split(testsuite)
    if dirname:
        os.chdir(dirname)
    mtime_s = os.stat(target).st_mtime
    mtime_c = os.stat(pyc).st_mtime if os.path.exists(pyc) else 0
    mtime_t = lastmod_of_testcase(testname)
    retcode = 0
    if mtime_s > mtime_c or max(mtime_s, mtime_c, mtime_t) == mtime_t:
        retcode = subprocess.call(['python', testname])
    if dirname:
        os.chdir(curdir)
    if retcode != 0 and os.path.exists(pyc):
        os.remove(pyc)
    return retcode

if __name__ == '__main__':
    args = sys.argv[1:]
    if len(args) >= 2:
        sys.exit(main(args[0], args[1]))
    elif len(args) == 1:
        sys.exit(main(args[0]))
    else:
        print 'tester.py PYTHON_SOURCE [TEST_SUITE]'
        sys.exit(1)

