from topdown import *
import types

def creole2html(text):
    html = mk_block_parser().run(normalize_lf(text) + '\n\n')
    return '\n'.join(map(Tag.output, html))


def escape(s):
    return s.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;').replace('"', '&quot;')

class Tag(object):
    def __init__(self, tag, attrs, node):
        self.tag = tag
        self.attrs = attrs
        self.childnode = node

    def output(self):
        if self.attrs:
            attr = u' '.join(map(lambda s:u'%s="%s"' % (s[0], escape(s[1])), self.attrs))
        else:
            attr = u''
        node = self._mk_node(self.childnode)
        if node:
            if attr:
                r = u'<%s %s>%s</%s>' % (self.tag, attr, node, self.tag)
            else:
                r = u'<%s>%s</%s>' % (self.tag, node, self.tag)
        else:
            if attr:
                r = u'<%s %s/>' % (self.tag, attr)
            else:
                r = u'<%s/>' % (self.tag)
        if self.tag in ('h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'ul', 'ol', 'li', 'table', 'tr', 'pre'):
            r = '\n%s\n' % r
        return r

    def _mk_node(self, node):
        if isinstance(node, list) or isinstance(node, types.GeneratorType):
            l = []
            for n in node:
                l.append(self._mk_node(n))
            return ''.join(l)
        else:
            if isinstance(node, basestring):
                return escape(node)
            else:
                if node:
                    return str(node)
                else:
                    return None

    def __str__(self):
        return self.output()

def mk_block_parser():
    return combine(*remove_line_feed(
                    header, 
                    horizonal,
                    orderedlist,
                    unorderedlist,
                    preformat,
                    table,
                    paragraph))

def mk_inline_parser():
    return combine(image,
                   hyperlink,
                   anchor,
                   strong,
                   emphasis,
                   linebreak,
                   plaintext)


def remove_line_feed(*components):
    def _remove_lf(f):
        def _wrapped(seq):
            r = f(seq)
            while seq.current == '\n':
                seq.next()
            return r
        _wrapped.__name__ = f.__name__
        _wrapped.__doc__ = f.__doc__
        _wrapped.__repr__ = lambda self:repr(f)
        return _wrapped
    return map(_remove_lf, components)

def header(seq):
    level = 0
    while seq.current == '=':
        seq.next()
        level += 1
    if not level:
        raise ParseFailed(seq, header)
    ws = seq.parse(' ')
    line = seq.parse(until('\n'))
    if level > 6: level = 6
    tag = 'h%d' % level
    return Tag(tag, None, [line.strip()])

def paragraph(seq):
    t = seq.parse(until('\n\n'))
    return Tag('p', None, [parse_inline(t)])

def list_block(seq, level, symbol):
    def list_item(seq):
        c = 0
        while seq.current == symbol:
            seq.next()
            c += 1
        if c != level:
            raise ParseFailed(seq, list_item)
        seq.parse(' ')
        ns = list('\n' + (symbol * level) + ' ')
        def cond(ls, xs):
            if len(ls) < len(xs):
                return True
            offset = len(ls) - len(xs)
            for i, v in enumerate(xs):
                if ls[offset + i] != v:
                    return True
            return False
        t = [seq.current]
        while cond(t, ns) and cond(t, ['\n', '\n']) and not seq.eof:
            c = seq.next()
            if c:
                t.append(c)
        if not cond(t, ns):
            for i in range(len(ns)):
                t.pop(-1)
            for i in range(len(ns)-1):
                seq.back()
        else:
            t.pop(-1)
            t.pop(-1)
            seq.back()
        t = ''.join(t)
        if not seq.eof:
            seq.next()
        t = ''.join(t)
        item = t
        nextlevel = level + 1
        nextsymbol = '\n' + symbol[0] * nextlevel
        subseq = CharSeq(item+u'\n\n')
        s = subseq.parse(until(nextsymbol))
        if subseq.eof:
            return Tag('li', None, [parse_inline(s)])
        else:
            subseq.next()
            f = lambda _s: list_block(_s, nextlevel, symbol)
            return Tag('li', None, [parse_inline(s), try_(as_parser(f))(subseq)])
    items = many(list_item)(seq)
    if not items:
        raise ParseFailed(seq, list_block)
    if symbol[0] == '*':
        tag = 'ul'
    elif symbol[0] == '#':
        tag = 'ol'
    return Tag(tag, None, items)

orderedlist = lambda seq: list_block(seq, 1, '#')
unorderedlist = lambda seq: list_block(seq, 1, '*')

def preformat(seq):
    _ = seq.parse('{{{')
    t = until('}}}')(seq)
    _ = seq.parse('}}}')
    return Tag('pre', None, [t])

def table(seq):
    if not seq.current == '|':
        raise ParseFailed(seq, table)
    return Tag('table', None, many1(table_row)(seq))

def table_row(seq):
    line = until('\n')(seq)
    lf = seq.parse('\n')
    return Tag('tr', None, many1(table_cell).run(line))

def table_cell(seq):
    _ = seq.parse('|')
    c = until('|')(seq)
    if not c:
        raise ParseFailed(seq, table_cell)
    h = c.startswith('=')
    if h:
        t = 'th'
        c = c[1:]
    else:
        t = 'td'
    return Tag(t, None, [parse_inline(c)])

def horizonal(seq):
    _ = seq.parse('----\n')
    return Tag('hr', None, None)

emphasis = (re.compile(r'(.*?)//(.+?)//(.*)'), lambda s, f: Tag('em', None, f(s)))
strong = (re.compile(r'(.*?)\*\*(.+?)\*\*(.*)'), lambda s, f: Tag('strong', None, f(s)))
image = (re.compile(r'(.*?)<<<img:(.+?)>>>(.*)'), lambda s, f: Tag('img', [('src', s)], None))
anchor = (re.compile(r'(.*?)\[\[(.+?\|.+?)\]\](.*)'), lambda s, f: (lambda x, y:Tag('a', [('href', x)], f(y)))(*s.split('|')))
url = (re.compile(r'(.*?)(https?://[-\w!@#$%^&*()_=+~:;\',./?]+)(.*)'), lambda s, f:Tag('a', [('href', s)], [s]))
patterns = (image, anchor, url, strong, emphasis)
def parse_inline(t):
    r = []
    t = t.replace('\n', ' ')
    for p, cls in patterns:
        m = p.match(t)
        if m:
            a = m.group(1)
            b = m.group(2)
            c = m.group(3)
            if a:
                r.append(parse_inline(a))
            r.append(cls(b, parse_inline))
            if c:
                r.append(parse_inline(c))
            break
    if not r:
        r.append(t)
    return r

