from topdown import *
"""
template = term*
term ::= string | statement
statement ::= var | if | for
var ::= '{' ws varref ws '}'
varref ::= normvar | rawvar
normvar ::= '$' 'a'..'z' ...
rawvar ::= '$' '!' 'a' .. 'z' ...
if ::= '{if' ws varref ws '}' template elif* else? '{end}'
elif ::= '{elif' ws varref ws '}' template
else ::= '{else}' template
for ::= '{for' ws name ws 'in' ws varref ws '}' template '{end}'
ws ::= ' '+
"""

class AST(object):
    def accept(self, visitor):
        raise NotImplemented

class Template(AST):
    def __init__(self, nodes):
        self.nodes = nodes

    def accept(self, visitor):
        r = []
        for n in self.nodes:
            r.append(n.accept(visitor))
        return ''.join(filter(lambda a: a != None, r))

class String(AST):
    def __init__(self, s):
        self.data = s

    def accept(self, vistor):
        return str(self)

    def __str__(self):
        return self.data

class If(AST):
    def __init__(self, varnode, subtempl, elifnodes, elsenode):
        self.varnode = varnode
        self.subtempl = subtempl
        self.elifnodes = elifnodes
        self.elsenode = elsenode

    def accept(self, visitor):
        v = self.varnode.accept(visitor)
        if v:
            return self.subtempl.accept(visitor)
        else:
            for elifnode in self.elifnodes:
                e = elifnode.accept(visitor)
                if e != None:
                    return e
            if self.elsenode:
                return self.elsenode.accept(visitor)
        return None

class Elif(AST):
    def __init__(self, varnode, subtempl):
        self.varnode = varnode
        self.subtempl = subtempl

    def accept(self, visitor):
        v = self.varnode.accept(visitor)
        if v:
            return self.subtempl.accept(visitor)
        else:
            return None

class Else(AST):
    def __init__(self, subtempl):
        self.subtempl = subtempl

    def accept(self, visitor):
        return self.subtempl.accept(visitor)

class For(AST):
    def __init__(self, itemname, varnode, subtempl):
        self.itemname = itemname
        self.varnode = varnode
        self.subtempl = subtempl

    def accept(self, visitor):
        loopvar = self.varnode.accept(visitor, True)
        visitor.add_context_layer()
        r = []
        for v in loopvar:
            visitor.put(self.itemname, v)
            r.append(self.subtempl.accept(visitor))
        visitor.delete_context_layer()
        return ''.join(r)

class VarRef(AST):
    def __init__(self, name):
        self.varname = name

    def accept(self, visitor, primitive=False):
        if primitive:
            return visitor.get_variable(self.varname)
        return escape(str(visitor.get_variable(self.varname)))

class RawVarRef(AST):
    def __init__(self, name):
        self.varname = name

    def accept(self, visitor):
        return str(visitor.get_variable(self.varname))

def escape(s):
    return s.replace('&', '&amp;').replace('"', '&quot;').replace('<', '&lt;').replace('>', '&gt;')

class Visitor(object):
    def __init__(self, context):
        self.context = context
        self.layer = []

    def get_variable(self, key):
        for l in self.layer:
            if key in l:
                return l[key]
        return self.context[key]

    def add_context_layer(self):
        self.layer.insert(0, {})

    def put(self, k, v):
        self.layer[0][k] = v

    def delete_context_layer(self):
        self.layer.pop(0)

def compile_template(t):
    return template.run(t)

@as_parser
def template(seq):
    t = seq.parse(many(term))
    return Template(t)

def term(seq):
    s = seq.parse([string, statement])
    return s

def string(seq):
    s = until(u'{')(seq)
    if not s:
        raise ParseFailed(seq, string)
    if s.endswith('\\'):
        s += seq.current
        seq.next()
    return String(s)

def statement(seq):
    s = seq.parse([if_, for_, try_(var)])
    return s

def var(seq):
    ob = seq.parse('{')
    ws = seq.parse(option(skip_ws))
    ref = varref(seq)
    ws = seq.parse(option(skip_ws))
    cb = seq.parse('}')
    return ref

def varref(seq):
    d = seq.parse('$')
    r = seq.parse(option('!'))
    v1 = seq.parse(alpha)
    v2 = seq.parse(option(alphanum, ''))
    if not r:
        return VarRef(v1 + v2)
    else:
        return RawVarRef(v1 + v2)

skip_ws = many(' ')

def if_(seq):
    _ = seq.parse('{if')
    ws = seq.parse(skip_ws)
    varnode = varref(seq)
    ws = seq.parse(skip_ws)
    _ = seq.parse('}')
    subtempl = template(seq)
    elifnodes = many(elif_)(seq)
    elsenode = option(else_)(seq)
    _ = seq.parse('{end}')
    return If(varnode, subtempl, elifnodes, elsenode)

def elif_(seq):
    _ = seq.parse('{elif')
    ws = seq.parse(skip_ws)
    varnode = var(seq)
    ws = seq.parse(skip_ws)
    _ = seq.parse('}')
    subtempl = template(seq)
    return Elif(varnode, subtempl)

def else_(seq):
    _ = seq.parse('{else}')
    subtempl = template(seq)
    return Else(subtempl)

def for_(seq):
    _ = seq.parse('{for')
    ws = seq.parse(skip_ws)
    v1 = seq.parse(alpha)
    v2 = seq.parse(option(alphanum, ''))
    itemname = v1 + v2
    ws = seq.parse(skip_ws)
    _ = seq.parse('in')
    ws = seq.parse(skip_ws)
    varnode = varref(seq)
    ws = seq.parse(skip_ws)
    _ = seq.parse('}')
    subtempl = template(seq)
    _ = seq.parse('{end}')
    return For(itemname, varnode, subtempl)


